unit LatinHypercubeRun;

interface

Uses SLR6;

Procedure UncertRun(SS: TSLAMM_Simulation);

implementation

Uses UncertDefn, CalcDist, RandNum, Global, TCollect, sysutils, dialogs, progress, Controls, forms,System.UITypes;

Const NUM_UNCERT_OUTPUTS = 27;

Procedure UncertRun(SS: TSLAMM_Simulation);
Type UResultsTypes = (DetRes,MinRes,MaxRes,StdRes,MeanRes);
Var ResLoop      : UResultsTypes;
    UResults     : Array of Array[DetRes..MeanRes] of Array of Double;
    Dist         : TInputDist;
    NumOutputFiles: Integer;
    NSLoop, DLoop: Integer;
    DrawValue    : Double;
    AllDataFileN : String;
    ErrorString  : String;
    TextOut      : TextFile;
    OutputFiles  : Array of TextFile;
    DateHolder   : String;
    FileBacked   : Array of String;
//    UserDist     : TUserDist;
    UserInterrupt: Boolean;
    IterationsDone: Integer;
   {------------------------------------------------------------------}
    Function NoCommas(InStr: String): String;
    Begin
      Result := InStr;
      While Pos(',',Result) > 0 do
        Result[Pos(',',Result)] := ':';
    End;
   {------------------------------------------------------------------}
    Function RandomInt(Top: Integer): Integer;
    {Returns a random integer from 1..Top, using Randnum.Pas}
    Begin
      RandomInt:=Trunc(RandUniform*Top)+1;
    End;
   {------------------------------------------------------------------}
    Function ICDF(Prob:Double): Double;
    Var Res: Double;
    Begin
     Res := Error_Value;
       With Dist do
        Case Disttype of
          Normal       : Res:=ICdfNormal(Prob,Parm[1],Parm[2]);
          Triangular   : Res:=ICdfTriangular(Prob,Parm[2],Parm[3],Parm[1]);
          LogNormal    : Res:=ICdfLogNormal(Prob,exp(Parm[1]),exp(Parm[2]));
          Uniform      : Res:=ICdfuniform(Prob,Parm[1],Parm[2]);
//          USER         : Res := UserDist.ICDF(Prob);

        end;
     If Res = Error_Value then Raise ESLAMMError.Create('Distribution Error!  ICDF Called with Invalid Parameters.');
     ICDF:=Res;
    End;
   {------------------------------------------------------------------}
    Function CDF(XVal:Double): Double;
    Var Res: Double;
    Begin
     Res := Error_Value;
       With Dist do
         Case Disttype of
               Triangular: Res:=cdfTriangular(XVal,Parm[2],Parm[3],Parm[1]);
               Normal:     Res:=cdfNormal(XVal,Parm[1],Parm[2]);
               LogNormal:  Res:=cdfLogNormal(XVal,exp(Parm[1]),exp(Parm[2]));
               Uniform:    Res:=cdfuniform(XVal,Parm[1],Parm[2]);
//               USER:       Res := UserDist.CDF(XVal);
          end;
     If Res = Error_Value then Raise ESLAMMError.Create('Distribution Error!  CDF Called with Invalid Parameters.');
     CDF:=Res;
    End;
   {------------------------------------------------------------------}
    Function CalculateDraw(Interval: Integer): Double;
   {This Procedure calculates the x value of the distribution
    given the interval in [1..NumIterations]}

    Var ProbLow,ProbHigh: Double;
        Width, Where: Double; {Describe Cum Probability Interval}
        RandProb: Double;     {Random Prob within Interval}
    Begin
      ProbHigh:=1.0;
      {Current implementation does not have truncated distributions
       so ProbHigh = 1.0 rather than CDF(Xmax)}

      {Xmin set to zero for now.}
      {triangular and uniform already are subject to Xmin of zero through
       the entry screen}
      ProbLow:=0.0;
      If (Dist.DistType in [normal,lognormal])
         then ProbLow:=CDF(0);

      {Width is the width of the probability interval, "Where" is the upper
       bounds of the probability interval, RandProb is the random value
       within the probability interval (From LATIN.FOR)}
      Width:=(ProbHigh-ProbLow)/SS.UncertSetup.Iterations;
      Where:=(Width*Interval) + Problow;
      RandProb:=Where - Width*RandUniform;

      CalculateDraw:=ICDF(RandProb);
    End;

   {------------------------------------------------------------------}
    Procedure FillVariableDraws(indx: Integer);
   {This Procedure fills the PCollection associated with each used
    distribution with NumSteps (number of iterations) UncertDraw objects
    each which has the numbered interval and the calculated draw from
    the Function CalculateDraw}

    {In order to calculate the rank order of each parameter's draw
    (ensures non-repeatability and tells which interval to sample from),
    the Procedure calculates a random number for each draw and then ranks
    them.  Optimized Feb 10, 97, JSC}

    Var IterationLoop: Integer;
        FindSlotLoop : Integer;
        LowValue     : Double;
        LowValIndex  : Integer;
        NewDraw      : TUncertDraw;

    Begin
      Dist.Draws:=New(PCollection,Init(500,50));

      {insert correct number of draws with random number that will
       then be ranked to get the Latin Hypercube Interval}
      With SS.UncertSetup do
       For IterationLoop:=1 to Iterations do
        begin
          If Indx>27 then NewDraw:= TUncertDraw.Init(0,IterationLoop/Iterations,0)
                     else NewDraw:= TUncertDraw.Init(0,RandUniform,0);
          Dist.Draws^.Insert(NewDraw);
        end;

{      UserDist := nil;
      If Dist.DistType = USER then
         UserDist := TUserDist.LoadFromFile(Dist.UserFileN); }

      With SS.UncertSetup do
       For IterationLoop:=1 to Iterations do
        Begin
          LowValue:=99;
          LowValIndex:=0;
          For FindSlotLoop:=0 to Iterations-1 do
              If TUncertDraw(Dist.Draws^.At(FindSlotLoop)).RandomDraw<=LowValue
                then
                  Begin
                    LowValIndex:=FindSlotLoop;
                    LowValue:=TUncertDraw(Dist.Draws^.At(FindSlotLoop)).RandomDraw;
                  End;
          NewDraw:=Dist.Draws^.At(LowValIndex);
          NewDraw.RandomDraw:=99;
          NewDraw.IntervalNum:=IterationLoop;
          NewDraw.Value:=CalculateDraw(IterationLoop)
        End;

 //     If UserDist <> nil then UserDist.Free;

    End;

   {------------------------------------------------------------------}

    Procedure Accumulate_Uncertainty_Results;
    Var OutLoop    : Integer;
        ResLp      : UResultsTypes;
        UncertVal, OutputVal: Double;

        {------------------------TOP-CREATEDISTS--------------------------}
        Procedure InitArrays;
        {Initialize the UResults Arrays}
        Var Loop, OFLoop : Integer;
            Res        : UResultsTypes;
            OVal       : Double;

        Begin
          For OFLoop := 0 to NumOutputFiles-1 do
           For Res:=MinRes to MeanRes do
             SetLength(UResults[OFLoop,Res],NUM_UNCERT_OUTPUTS);

          For OFLoop := 0 to NumOutputFiles-1 do
           For Loop :=0 to NUM_UNCERT_OUTPUTS-1 do  {for each datapoint}
             For Res := MinRes to MeanRes do
                  Begin
                    OVal :=  SS.Summary[OFLoop,SS.UncertSetup.UncSens_Row-1, Loop ];
                    If Res=MinRes then {avoid writing first set of raw data multiple times}
                      Write(OutputFiles[OFLoop],FloatToStrF(OVal,FFExponent,12,2)+', ');

                    If Res=StDRes then UResults[OFLoop,Res,Loop] := sqr(OVal)
                                  else UResults[OFLoop,Res,Loop] := OVal;
                  End;
        End;
        {---------------------BOTTOM-CREATEDISTS--------------------------}

    Var i, OFLoop: Integer;
        OutputFileN, OFString: String;

    Begin    {Accumulate_Uncertainty_Results}
      SetLength(OutputFiles,NumOutputFiles);

      If NSLoop=1 then with SS.UncertSetup do
       For OFLoop := 0 to NumOutputFiles-1 do
        Begin
          OFString := '';
          If OFLoop > 0 then OFString := '_OS'+IntToStr(OFLoop);
          If OFLoop > SS.Site.NOutputSites then OFString := '_ROS'+IntToStr(OFLoop-SS.Site.NoutputSites);

          OutputFileN := ChangeFileExt(AllDataFileN,OFString+'_alldata.csv');
          While Pos('\\',OutputFileN)>0 do
            Delete(OutputFileN,Pos('\\',OutputFileN),1);

          ASSIGNFILE(OutputFiles[OFLoop],OutputFileN);
          REWRITE(OutputFiles[OFLoop]);

          For i:=0 to NumDists-1 do
             Write(OutputFiles[OFLoop],'"Mult. '+NoCommas(Distarray[i].Name)+'", ');

          For i:=0 to NUM_UNCERT_OUTPUTS-1 do
            Write(OutputFiles[OFLoop],'"'+NoCommas(SS.ColLabel[i] )+'", ');
        End;  //output file loop, file initialization

      For OFLoop := 0 to NumOutputFiles-1 do
       Begin
        Writeln(OutputFiles[OFLoop]);
        With SS.UncertSetup do
        For i:=0 to NumDists-1 do
         If DistArray[i].DistType <> ElevMap
           then Write(OutputFiles[OFLoop],FloatToStrF(TUncertDraw(Distarray[i].Draws^.At(NSLoop-1)).Value,FFExponent,12,2)+', ')
           else Write(OutputFiles[OFLoop],DistArray[i].Name,' , ');
       End; {OFLoop, write distributions}

      If NSLoop=1 then InitArrays else
        Begin
         For OFLoop := 0 to NumOutputFiles-1 do
          For OutLoop :=0 to NUM_UNCERT_OUTPUTS-1 do  {for each datapoint}
            Begin
              OutputVal := SS.Summary[OFLoop,SS.UncertSetup.UncSens_Row-1,OutLoop ];
              Write(OutputFiles[OFLoop],FloatToStrF(OutputVal,FFExponent,12,2)+', ');
              For ResLp:=MinRes to MeanRes do
                Begin
                  UncertVal := UResults[OFLoop,ResLp,OutLoop];
                  Case ResLp of
                    MeanRes: UResults[OFLoop,ResLp,OutLoop] := UncertVal+OutputVal;
                             {Sum data for now}
                    MinRes : If OutputVal<UncertVal then
                               UResults[OFLoop,ResLp,OutLoop] := OutputVal;
                    MaxRes : If OutputVal>UncertVal then
                               UResults[OFLoop,ResLp,OutLoop] := OutputVal;
                    StdRes : UResults[OFLoop,ResLp,OutLoop]   := UncertVal + (OutputVal*OutputVal);
                            {Sum of Squares for now}
                  End; {Case}
                End; {ResLp}
            End; {OutLoop}
        End; {else}

      {Update Text output}
      IterationsDone:=NSLoop;
      Writeln(Textout);
      Writeln(Textout,'Iteration '+IntToStr(NSLoop)+' Completed.');
      Writeln(Textout,'---------------------------------------------------------');

    End;      {Accumulate_Uncertainty_Results}
   {------------------------------------------------------------------}
    Function InitTextResults:Boolean;
    Var FileN: String;
        curdir: String;
    Begin
      CurDir := GetCurrentDir;
      Result := PromptForFileName(SS.UncertSetup.CSV_path,'CSV File (*.csv)|*.csv','.csv','Select Base Output File',SS.NWIFileN ,True);
      SetCurrentDir(CurDir);

      If Not Result then Begin UserInterrupt:=True; SS.UserStop:=True; Exit; End;

      FileN := ChangeFileExt(SS.UncertSetup.CSV_path,'_uncertlog.txt');

      While Pos('\\',FileN)>0 do
        Delete(FileN,Pos('\\',FileN),1);

      ASSIGNFILE(TextOut,FileN);
      REWRITE(TextOut);
      Writeln(Textout,'---------------------------------------------------------');
      Writeln(TextOut);
      Writeln(TextOut,'        Uncertainty Run for SLAMM                        ');
      Writeln(TextOut);
      Writeln(Textout,'---------------------------------------------------------');
      DateTimeToString(DateHolder,'mm-d-y t',Now);
      Writeln(TextOut,'        Run Starts at '+ DateHolder);
      Writeln(TextOut,'---------------------------------------------------------');
      Writeln(TextOut);
      Writeln(TextOut,'        Technical Details for Uncert Viewer              ');
      Writeln(TextOut);
      Writeln(TextOut,'~FILENAME=',ExtractFileName(SS.UncertSetup.CSV_Path));
      Writeln(TextOut,'~NUM_ITER=',SS.UncertSetup.Iterations);
      Writeln(TextOut,'~TIME_ZERO=',SS.Site.T0);
      Writeln(TextOut);
      Writeln(TextOut,'---------------------------------------------------------');
      Writeln(TextOut);
      Writeln(TextOut,'        ** DISTRIBUTIONS SUMMARY **');
      Writeln(TextOut);
      Writeln(TextOut,'     All Distributions are Multipliers');
      Writeln(TextOut);

     End;
   {-------------------------------------------------------------------}
    Procedure DistToText;
    {Summarizes the distribution for the text output}
    Begin
    With Dist do
      Begin

        Writeln(TextOut,Name,':   Point Estimate (global):'+FloatToStrF(GetGlobalValue ,ffGeneral,4,4));
        Writeln(TextOut,SummarizeDist);
        Writeln(TextOut);
      End; {with}
    End; {Proc}
   {------------------------------------------------------------------}
    Procedure WriteDistSummaries;
   {------------------------------------------------------------------}
       Procedure PostProcessResults;
       { calculate stdev and mean from sum and sumsquare data }
       Var i,j : Integer;
           Sum,SumSquare: Double;
           InSqrt     : Double;
           n          : Integer;
       Begin
        For j := 0 to NumOutputFiles-1 do
         For i:=0 to NUM_UNCERT_OUTPUTS-1 do   {For every datapoint}
           Begin
              Sum       := UResults[j,MeanRes,i];
              SumSquare := UResults[j,StdRes,i];
              n:=IterationsDone;

              If n>0 then UResults[j,MeanRes,i]:=Sum/n;
              {The standard deviation is calculated using the "nonbiased" or "n-1" method.}
              If n>1 then InSqrt:= ((n*SumSquare)-(Sum*Sum)) / (n*(n-1))
                     else InSqrt:=0;
              If InSqrt>0 then UResults[j,StdRes,i] :=Sqrt(InSqrt)
                          else UResults[j,StdRes,i] :=0;
          End; {i loop}
       End; {PostProcessResults}

   {----------------- Procedure WriteDistSummaries  ---------------}
   Var OFLoop, i : Integer;

       Begin
         If IterationsDone>0 then PostProcessResults;

         For OFLoop := 0 to NumOutputFiles-1 do
           Begin
             Writeln(OutputFiles[OFLoop]);
             Writeln(OutputFiles[OFLoop]);
             Writeln(OutputFiles[OFLoop], 'Variable Name, Min, Mean, Max, Std. Dev., Deterministic');

              { Loop through outputvariables and write results to Excel }
                For i:=0 to NUM_UNCERT_OUTPUTS-1 do   {For every datapoint}
                 Begin
                  Writeln(OutputFiles[OFLoop], SS.ColLabel[i],',',UResults[OFLoop,minRes,i],',',UResults[OFLoop,meanRes,i],',',UResults[OFLoop,maxRes,i],
                                      ',',UResults[OFLoop,StdRes,i],',',UResults[OFLoop,DetRes,i]);
                 End; {For i}
           End;
       End;  {Procedure WriteDistSummaries}

   {------------------------------------------------------------------}

       Function VerifyUncertSetup: Boolean;
       Var Loop, FixLoop, CountRuns: Integer;
           RunningCustomSLR: Boolean;
           IPCC_Scenario: IPCCScenarios;
           IPCC_Est     : IPCCEstimates;
           Prot_Scenario: ProtScenario;
           ErrMessage   : String;

       Begin
         Result := True;
         CountRuns := 0;

         With SS do
           Begin
            For IPCC_Scenario := Scen_A1B to Scen_B2 do
              For IPCC_Est := Est_Min to Est_Max do
               For Prot_Scenario := NoProtect to ProtAll do
                If IPCC_Scenarios[IPCC_Scenario] and IPCC_Estimates[IPCC_Est] and Prot_To_Run[Prot_Scenario] then
                 Inc(CountRuns);

             For Prot_Scenario := NoProtect to ProtAll do
               For FixLoop := 1 to 3 do
                 If Fixed_Scenarios[FixLoop] and Prot_To_Run[Prot_Scenario] then
                  Inc(CountRuns);

             RunningCustomSLR := False;
             If RunCustomSLR then
               For Prot_Scenario := NoProtect to ProtAll do
                 If Prot_To_Run[Prot_Scenario] then
                  Begin
                   Inc(CountRuns);
                   RunningCustomSLR := True;
                  End;
           End;

         ErrMessage := '';
         If (CountRuns<>1) then ErrMessage := 'Uncertainty runs must be set up to run one SLR scenario / Protection Scenario only.';
         If Not RunningCustomSLR then
           Begin
              For Loop:=0 to SS.UncertSetup.NumDists-1 do
                 Begin
                   Dist:=SS.UncertSetup.DistArray[Loop];
                   If (Dist.IDNum = 1) then ErrMessage := 'You must run SLAMM under "Custom SLR" if you are selecting the "Sea Level Rise by 2100 (multiplier)" as an uncertainty distribution.';
                 End;
           End;

         If ErrMessage <> '' then
           Begin
             Result := False;
             MessageDlg(ErrMessage,MtError,[mbok],0);
             SS.UserStop := True;
           End;
       End;

   {------------------------------------------------------------------}

      Procedure RestoreVals;
      Var DLoop:Integer;
      Begin
       With SS.UncertSetup do
        For DLoop:=0 to NumDists-1 do
          Begin
            Dist:=DistArray[DLoop];
            If Dist.DistType <> ElevMap then Dist.RestoreValues;
         End;
       End;

  {----------------------------------------------------------------}
  {---        BEGIN MAIN Procedure LATIN HYPERCUBE RUN          ---}
  {----------------------------------------------------------------}
Var i, OFLoop, Loop,Prog: Integer;
    OVal: Double;
    ItStr: String;
    TestFileN: String;
Begin  {UncertRun}
  IterationsDone:=0;
  UncAnalysis := True;
  SS.UncertSetup.UncSens_Iter := 0;
  SS.WordInitialized := False;
  UserInterrupt:=False;
  SS.UserStop := False;

//  If Not VerifyUncertSetup then exit;    {Marco: already checked when clicking execute button}

  Try

  If not InitTextResults then Exit;

  AllDataFileN := SS.UncertSetup.CSV_path;
  TestFileN := ChangeFileExt(AllDataFileN,'_alldata.csv');
  While Pos('\\',TestFileN)>0 do
     Delete(TestFileN,Pos('\\',TestFileN),1);
  If FileExists(TestFileN) then
   If MessageDlg('Overwrite '+TestFileN+'?',mtconfirmation,[mbyes,mbno],0) = mrno then
     if not PromptForFileName(AllDataFileN,'CSV File (*.csv)|*.csv','.csv','Select AllData Output File',TestFileN,True)
       then Begin UserInterrupt:=True; SS.UserStop:=True; Exit; End;

  If SS.UncertSetup.UseSeed then SetSeed(SS.UncertSetup.Seed)
                            else SetSeed(-1);


  ProgForm.ModalResult := MRNone;
  ProgForm.Gauge1.Progress:=0;
  ProgForm.Show;
  ProgForm.UncertStatusLabel.Visible := True;
  ProgForm.UncertStatusLabel.Caption := 'Deterministic Run...';
  ProgForm.Update;
  For i := 1 to 2 do
    SS.UncertSetup.ZUncertMap[i] := nil;

  Try
    SS.ExecuteRun;
  Except
     Messagedlg('Model Execution Error.',mterror,[mbok],0);
     CloseFile(TextOut);
     ProgForm.Hide;
     exit;
  End;

  If SS.UserStop then
    Begin
      CloseFile(TextOut);
      ProgForm.Hide;
      exit;
    End;

  NumOutputFiles := 1+SS.Site.NOutputSites+SS.Site.MaxROS;
  SetLength(UResults,NumOutputFiles);

  For OFLoop := 0 to NumOutputFiles-1 do
    Begin
      SetLength(UResults[OFLoop,DetRes],NUM_UNCERT_OUTPUTS);
      For Loop:=0 to NUM_UNCERT_OUTPUTS-1 do  {for each datapoint}
        Begin
          OVal :=  SS.Summary[OFLoop,SS.UncertSetup.UncSens_Row-1, Loop];
          UResults[OFLoop,DetRes,Loop] := OVal;   //save deterministic results
        End;
    End;

  FileBacked := nil;

  ProgForm.UncertStatusLabel.Caption := 'Calculating Latin Hypercube Draws...';
  ProgForm.Update;

  {The below loop fills the VariableDraws and saves
   the point estimates so they can be restored after the LH run}
  For Loop:=0 to SS.UncertSetup.NumDists-1 do
     Begin
       Dist:=SS.UncertSetup.DistArray[Loop];
       If Dist.DistType <> ElevMap then FillVariableDraws(Loop);
       DistToText;
     End;

  Writeln(TextOut,'---------------------------------------------------------');

  For i := 1 to 2 do
    SS.UncertSetup.ZUncertMap[i] := nil;
  With SS.UncertSetup do
   For NSLoop:=1 to Iterations do
    Begin
      {Update Progress Dialog}
      ItStr:='Iteration '+IntToStr(NSLoop+GIS_Start_Num-1)+ ' of '+IntToStr(Iterations+GIS_Start_Num-1);
      ProgForm.UncertStatusLabel.Caption := ItStr;
      Writeln(TextOut,Itstr);
      Writeln(TextOut);

      {Load the latin hypercube values into the simulation}
      For DLoop:=0 to NumDists-1 do
        Begin
          Dist:=DistArray[DLoop];
          With Dist do
           If DistType <> ElevMap
             then
                Begin
                  DrawValue:=TUncertDraw(Draws^.At(NSLoop-1)).Value;
                  SetValues(DrawValue);
                  Writeln(TextOut,Dist.Name,' (Multiplied by) '+FloatToStrF(DrawValue,ffGeneral,6,5));
                End
             else Begin
                    SS.UncertSetup.MakeUncertMap(Dist.ZMapIndex, SS.Site.Rows,SS.Site.Cols,Dist.Parm[1],Dist.Parm[2]);  //write spatially autocorrelated elevation uncertainty map
                    Writeln(TextOut,Dist.Name,'  RMSE '+FloatToStrF(Dist.Parm[1],ffGeneral,6,5));
                  End;
       End;

       Inc(UncSens_Iter);

        {Go through run Procedure for each iteration}
        Try
          SS.ExecuteRun;
        Except
          RestoreVals;
          UserInterrupt:=True;
          SS.UserStop := True;

          DateTimeToString(DateHolder,'mm-d-y t',Now);
          Writeln(TextOut,'Execute Model Error at '+ DateHolder );
          Messagedlg('Model Execution Error.',mterror,[mbok],0);
          For i := 1 to 2 do
            SS.UncertSetup.ZUncertMap[i] := nil;
          Break; {exit distribloop}
        End;

      RestoreVals;

      Prog := Round ((NSLoop/Iterations) * 100);
        If Prog <> ProgForm.Gauge1.Progress then
          Begin
            ProgForm.Gauge1.Progress := Prog;
            ProgForm.Update;
          End;  {If Round}

      Application.ProcessMessages;
      Application.Title := 'SLAMM6 - '+IntToStr(nsloop)+' of '+IntToStr(Iterations);

      If ProgForm.ModalResult<>0 then
        Begin
          UserInterrupt:=True;
          SS.UserStop := True;
        End; {user interrupt}

      If (ProgForm.ModalResult<>0) or SS.UserStop then
         begin
            UserInterrupt:=True;
            SS.UserStop := True;
            DateTimeToString(DateHolder,'mm-d-y t',Now);
            Writeln(TextOut,'Run Terminated at '+ DateHolder );
            Break; {exit loop}
         end
         else Accumulate_Uncertainty_Results;
    End;  {NSLoop}

    Application.Title := 'SLAMM 6 beta';
    If Not UserInterrupt and not SS.UserStop then
      begin
        DateTimeToString(DateHolder,'mm-d-y t',Now);
        Writeln(TextOut,'Run Successfully Completed At '+ DateHolder );
        Writeln(TextOut,'---------------------------------------------------------');
      end;

  Except
    UncAnalysis := False;
    For i := 1 to 2 do
      SS.UncertSetup.ZUncertMap[i] := nil;

    ErrorString:=Exception(ExceptObject).Message;
    ProgForm.ModalResult:=1;
    MessageDlg('Run-Time Error During Uncertainty Iteration',mterror,[mbOK],0);
    MessageDlg(ErrorString,mterror,[mbOK],0);
    DateTimeToString(DateHolder,'mm-d-y t',Now);
      Try
        Writeln(TextOut,'Run Terminated at '+ DateHolder );
        Writeln(TextOut,'    Due to '+ErrorString);
      Except
        MessageDlg('No Data Written',mterror,[mbOK],0);
        ProgForm.ModalResult:=1;
      End;

  End; {Except}

  UncAnalysis := False;
  For i := 1 to 2 do
    SS.UncertSetup.ZUncertMap[i] := nil;

  Try

   IF IterationsDone>0 then WriteDistSummaries;

   For OFLoop := 0 to NumOutputFiles-1 do
     For ResLoop:=DetRes to MeanRes do
       Finalize(UResults[OFLoop,ResLoop]);
   Finalize(UResults);

  CloseFile(TextOut);
  For OFLoop := 0 to NumOutputFiles-1 do
    CloseFile(OutputFiles[OFLoop]);

  Except
    ErrorString:=Exception(ExceptObject).Message;
    MessageDlg('Run-Time Error Writing to File After Uncertainty Run',mterror,[mbOK],0);
    MessageDlg(ErrorString,mterror,[mbOK],0);
    ProgForm.ModalResult:=1;

    CloseFile(TextOut);
    For OFLoop := 0 to NumOutputFiles-1 do
      CloseFile(OutputFiles[OFLoop]);

  End; {Except}

  ProgForm.Hide;
End;  {LatinHyperCubeRun}



end.
