//SLAMM SOURCE CODE copyright (c) 2009 - 2012 Warren Pinnacle Consulting, Inc.
//Code Use and Redistribution is Subject to Licensing, SEE SLAMM_License.txt
//

UNIT UTILITY;

    INTERFACE

USES GLOBAL;

    (***************************************)
    (*   utilitiy procedures               *)
    (***************************************)

FUNCTION GetCellCat(Cl: PCompressedCell): Category;
FUNCTION CellWidth(Cl: PCompressedCell; Cat: Category): Single;
FUNCTION CatElev(Cl: PCompressedCell; Cat: Category): Single;
PROCEDURE SETCatElev(Cl: PCompressedCell; Cat: Category; SetVal: Single);
PROCEDURE SETCELLWIDTH(Cl: PCompressedCell; Cat: Category; SetVal: Single);
Function GetMinElev(Cl: PCompressedCell): Double;
//Function BELOWSALTBOUND(Cl: PCompressedCell): Boolean;
Function OPENWATER(Cl: PCompressedCell): Boolean;

  IMPLEMENTATION

{***************************************************
*                                                  *
*     GetCellCat: Determines the cell category     *
*       based on the percentages of classes.       *
*                                                  *
*   2/11/88       Written by - Manjit S. Trehan    *
*   3/13/88       Modified to Function by R. Park  *
***************************************************}

FUNCTION GetCellCat(Cl: PCompressedCell): Category;

VAR
  i: Integer;          { Loop Variable }
  CellCat : Category;
  MaxArea,
  CatArea : double;   { Percentage of the dominant class }

BEGIN { GetCellCat }
  MaxArea := 0.0;
  CellCat := Blank;
  FOR i := 1 to NUM_CAT_COMPRESS DO
    Begin
      CatArea := Cl.Widths[i];
      IF CatArea > MaxArea THEN
        BEGIN
          CellCat := Cl.Cats[i];
          MaxArea := CatArea;
        END;
    End;
   GetCellCat := CellCat;
END; { GetCellCat }


Function CellWidth(Cl: PCompressedCell; Cat: Category): Single;
Var i: Integer;
Begin
  CellWidth := 0;
  FOR i := 1 to NUM_CAT_COMPRESS DO
    Begin
      If Cl.Cats[i] = Cat then
        Begin
          CellWidth := Cl.Widths[i];
          Exit;
        End;
    End;
End;

{Function BELOWSALTBOUND(Cl: PCompressedCell): Boolean;
begin
   Cl.
end; }

Function OPENWATER(Cl: PCompressedCell): Boolean;
var water_width: double;
begin
  water_width:= (CellWidth(cl,OpenOcean) + CellWidth(cl,EstuarineWater) + CellWidth(cl,RiverineTidal)+ CellWidth(cl,TidalCreek));
  result := water_width > 0;
end;

Function CatElev(Cl: PCompressedCell; Cat: Category): Single;
Var i: Integer;
Begin
  CatElev := 99999;
  FOR i := 1 to NUM_CAT_COMPRESS DO
    Begin
      If Cl.Cats[i] = Cat then
        Begin
          CatElev := Cl.MinElevs[i];
          Exit;
        End;
    End;
End;

PROCEDURE SetCatElev(Cl: PCompressedCell; Cat: Category; SetVal: Single);
Var i: Integer;
Begin
  FOR i := 1 to NUM_CAT_COMPRESS DO
    Begin
      If Cl.Cats[i] = Cat then
        Begin
          Cl.MinElevs[i] := SetVal;
          Exit;
        End;
    End;
End;

PROCEDURE SETCELLWIDTH(Cl: PCompressedCell; Cat: Category; SetVal: Single);
Var i: Integer;
    MinWidth: Double;
Begin
  MinWidth := 99999;
  FOR i := 1 to NUM_CAT_COMPRESS DO
    Begin
      If Cl.Cats[i] = Cat then
        Begin
          Cl.Widths[i] := SetVal;
          Exit;
        End;
      If MinWidth> Cl.Widths[i] then MinWidth := Cl.Widths[i];
    End;

  If MinWidth > 0 then
    Begin
      Cat := Cat;  {Note debug point for test of data loss}
    End;

  {Apparently Category wasn't Found8}
  FOR i := 1 to NUM_CAT_COMPRESS DO
    Begin
      If Cl.Widths[i] = MinWidth then
        Begin
          If MinWidth < SetVal then {keep maximum of two minimum categories}
            Begin
              Cl.Cats[i] := Cat;
              Cl.MinElevs[i] := 999;
            End;

          Cl.Widths[i]   := SetVal + MinWidth;
          Exit;
        End;
    End;



  Raise ESLAMMError.Create('Algorithm Error, SETCELLWIDTH, min width not found');

End;


Function GetMinElev(Cl: PCompressedCell): Double;
Var MnElev: Double;
    i: Integer;
Begin
  MnElev := 9999999;
  For i := 1 to NUM_CAT_COMPRESS do
    If (CL.MinElevs[i]<MnElev) and (Cl.Widths[i] > 0) then
        MnElev := CL.Minelevs[i];
  GetMinElev := MnElev;

End;


End.
