//SLAMM SOURCE CODE copyright (c) 2009 - 2012, 2010 Warren Pinnacle Consulting, Inc.
//Code Use and Redistribution is Subject to Licensing, SEE SLAMM_License.txt
//

(*****************************************************)
(*                                                   *)
(*    SLAMM - SEA LEVEL AFFECTING MARSHES MODEL      *)
(*               Version 6 BETA                      *)
(*                                                   *)
(*  Jonathan S. Clough,                              *)
(*  Warren Pinnacle Consulting, Inc.                 *)
(*                                                   *)
(*  Richard A. Park   Eco Modeling                   *)
(*                                                   *)
(*  Version 3 by Park, Manjit Trehan, and Jae Lee    *)
(*                                                   *)
(*****************************************************)
                                                                                                     
Unit SLR6;

interface

Uses

  SysUtils,Utility, Windows, Messages, Classes, Graphics, Controls, Global, Uncert,
  Forms, Dialogs, StdCtrls, Math, Progress, GISOptions, OLECtrls, ComObj, stack, SalArray;

CONST
  Output_maxc = 48;

  SaltWater   : BYTE = 1;
  ExposedWater: BYTE = 2;
  MostlyWater : BYTE = 4;
  AllWetland  : BYTE = 8;

  Small = 0.1;  {sq m}

TYPE

  MapDisk    = ARRAY OF Integer;
  MapWord    = Array of Word;
  MapBoolean = ARRAY of BYTE;
  PctArray   = ARRAY[Category] OF double;
  ErosionScale = (None,Little,Moderate,Heavy,Severe);


TYPE
  OutputArray    = Array of Array of Array[0..Output_maxc] OF double;  {Output [0..NumSites+1,0..NumRows-1,Columns]}
  str_vector     = Array[0..Output_maxc] of string; {[25]; }
  str_vector1    = Array of string; {[25]; }
  file_name_type = String;


 TSLAMM_Simulation = Class
     { Object oriented class is far too flat for optimal design but remains a function of
       previous non-OOP code structure }
     {-------------------------------------------------------------}
     FileN, SimName, Descrip    : String;

     Site                : TSite;
     FWFlows             : TFWFlows;
     NumFWFlows          : Integer;
     SalRules            : TSalinityRules;

     TimeStep            : Integer;
     MaxYear             : Integer;
     RunSpecificYears    : Boolean;
     YearsString         : String;
     RunUncertainty      : Boolean;
     RunSensitivity      : Boolean;
     UncertSetup         : TSLAMM_Uncertainty;

     Fixed_Scenarios     : Array[1..3] of Boolean;
     IPCC_Scenarios      : Array[IPCCScenarios] of Boolean;
     IPCC_Estimates      : Array[IPCCEstimates] of Boolean;
     RunCustomSLR        : Boolean;
     CustomSLR           : Double;
     Prot_To_Run         : Array[ProtScenario] of Boolean;

     Make_Data_File      : Boolean;
     Display_Screen_Maps : Boolean;
     QA_Tools            : Boolean;
     RunFirstYear        : Boolean;
     Maps_to_MSWord      : Boolean;
     Maps_to_GIF         : Boolean;
     ExtraMaps           : Boolean;
     DucksMaps           : Boolean;
     SaveGIS             : Boolean;
     SaveElevsGIS, SaveSalinityGIS   : Boolean;
     BatchMode           : Boolean;

     CheckConnectivity   : Boolean;
     UseSoilSaturation   : Boolean;
     LoadBlankIfNoElev   : Boolean;

     UseBruun            : Boolean;
     InitZoomFactor      : Double;

     SaveMapsToDisk      : Boolean;
     MakeNewDiskMap      : Boolean;
     IncludeDikes        : Boolean;

     ClassicDike         : Boolean;

     ElevRanges          : TClassElevArray;
     ElevStats           : TElevStats;

     GridColors          : GridColorType;
     ElevGridColors      : ElevGridColorType;

    {-- File Name Management}
     BatchFileN          : String;
     ElevFileN, NWIFileN, SLPFileN, OutputFileN, {main files}
     IMPFileN,  ROSFileN, DikFileN, VDFileN, UpliftFileN, SalFileN : String;

     OptimizeLevel : Word;  {0 = no optimization, 1 = blank spaces not tracked, 2 = black, ocean, high dry land}
     NumMMEntries: Integer;

     {-------------------------------------------------------------}
     {NO LOAD OR SAVE BELOW}

     { Relevant to Current Simulation Run, no load save Reqd. }
       GridColors2          : GridColor2Type;

       Changed    : Boolean;
       CountIter  : Integer;
       CellHa     : double;

       FracLost : PctArray;

       Year: INTEGER;

       Tropical,            {TRUE if mangroves are present}

       {Has context only when examining a specific cell}
       ClearPathToOcean,    {True if 500 m to off-shore there's nothing but water and AdjOcean}
       AdjOcean,            {TRUE if within 500 m of open ocean, off-shore}
       AdjWater,            {TRUE if within 500 m of water}
       AdjSalt: Boolean;    {TRUE if within 500 m of saltwater wetland}

       Erosion: ErosionScale;   {degree of erosion on a qualitative scale}

       ProtectDeveloped,    {TRUE if developed areas are to be protected}
       ProtectAll: Boolean; {TRUE if all dry lands are to be protected}
       Running_Fixed: Boolean;
       Running_Custom: Boolean;
       IPCCSLRate: IPCCScenarios;
       IPCCSLEst: IPCCEstimates;
       FixedScen: Integer;

       MaxMTL, MinMTL    : single;

       Map               : MSArray;    {Collection of all cells in RAM }
       S60FileN          : String;     {Name of map on disk}
       MapHandle         : THandle;
       BackupFN          : String;
       DikeInfo          : TDikeInfoArr;
       WordApp, WordDoc  : Variant;
       WordInitialized   : Boolean;
       Summary           : OutputArray; {Output [0..NumSites+1,0..NumRows-1,0..MaxColumns]}
       ColLabel          : str_vector;
       FileStarted       : Boolean;      { batch CSV file Started? }
       RowLabel          : str_vector1;
       BMatrix : MapBoolean;
       ErodeMatrix: MapWord;        {Matrix of Erosion that has already occured at a given cell coordinate}
       MapMatrix: MapDisk;
       DataElev : Array of Word;    {Elevations  0..65535 in mm}
       MaxFetchArr : Array of Word; {Max Fetch   0..65535 in m}
       ROSArray : Array of Byte;    {Raster Output Site Relevant for Each Cell  0..255}
       CatSums : Array of PctArray;
       BlankCell: CompressedCell;
       UndDryCell, DevDryCell, OceanCell, EOWCell : CompressedCell;
       FRs : Array of FractalRecord;
       Init_ElevStats      : Boolean;  {have elevation stats been initialized?}
       UserStop            : Boolean;

       FracFreshFileN,SiteFileN,SubSiteFileN,AttribFileN: String;  {for reading SLAMM 5 files}

       SalArray      : TSalArray;
       SalStats      : TSalStats;
       Init_SalStats : Boolean;

       ROS_Resize : Integer;

       Connect_Array : Array of byte;                  // Unchecked = 1
                                                       // Above Salt Bound = 2
                                                       // Checked & Connected = 3
                                                       // Checked & Unconnected = 4
                                                       // Irrelevant = 5
                                                       // Diked or Blank = 7
                                                       // {Tidal Water = 8 [navy]}
                                                       // (Fashion after "DataElev" or "MaxFetchArr" in terms of initialization and usage)

     {NO LOAD OR SAVE ABOVE}
     {-------------------------------------------------------------}
     Constructor Create;
     Constructor Load(FN: string); overload;
     Procedure Save(FN: string); overload;
     Constructor Load(TS: TStream); overload;
     Procedure Save(var TS: TStream); overload;
     Destructor Destroy; override;
     {-------------------------------------------------------------}

     Function  Calc_Connectivity(testR,testC: Integer) :Boolean;

     Procedure Count_MMEntries;
     Procedure CheckForAnnualMaps;
     Procedure ExecuteRun;
     Procedure Calc_Elev_Stats;  {run elev analysis and copy results into ELEVStats}
     Procedure Calc_Elev_Stats_2;
     Function  MakeFileName:String;
     Function Save_GIS_Files: Boolean;
     Procedure Save_CSV_File;
     Function CopyMapsToMSWord: Boolean;
     Function MakeDataFile(CountOnly: Boolean; YearDikeFName, YearSalFName: String): Boolean;
     Procedure SaveRaster(FileN: String; rasterType: Integer);
     PROCEDURE SetMapAttributes;
     Procedure CalibrateSalinity;
     Procedure Calc_Fractal(Var Site:TSite);
     Function RunOneYear(IsT0: Boolean): Boolean;
     Procedure FractalNos(ER,EC: Integer; Var Cell:CompressedCell; Var FSite: TSite);
     Function ValidateOutputSites: Boolean;
     Procedure AssignGridColors2;

     {TRANSFER.INC: PROCEDURES THAT HANDLE SLR, INUNDATION, EROSION, OVERWASH in TRANSFER.INC}
     Function DynamicAccretion(Cell:PCompressedCell; SS: TSubsite; ModelNum: Integer): Double;  {calculated accr, mm/year}
     Function RegFloodAccrete(Cell:PCompressedCell; SS: TSubsite): Double;   {accretion rates, mm / year}
     Function IrregFloodAccrete(Cell:PCompressedCell; SS: TSubsite): Double; {accretion rates, mm / year}
     Function TidalFreshAccrete(Cell:PCompressedCell; SS: TSubsite): Double; {accretion rates, mm / year}
     Function TidalFlat_Sed(Cell:PCompressedCell; SS: TSubsite): Double;  {accretion rates, mm / year}

     Procedure LocalSLChange(VAR Site: TSite; VAR Year: Integer; IsT0: Boolean);
     Procedure TRANSFER(Cell: PCompressedCell; EachRow,EachCol: Integer; ErodeStep: Boolean);
     Function UpdateElevations: Boolean;  {update cell elevations for entire map based on SLR, uplift, subsidence, accretion}
     Function ChngWater(CalcMF, CalcGeometry: Boolean):Boolean;
     Function CalculateMaxFetch: Boolean;

     {END TRANSFER.INC}

     {INITX.INC: PROCEDURES RUN EARLY IN SLAMM SIMULATION, OR ELEV BOUNDARIES FUNCS CONTAINED IN INITX.INC}
     Procedure SetDefaultCategoryVariables;
     Function PreprocessWetlands:Boolean; {Elevation Processor for low quality elevation data}
     Procedure InitializeDikes;
     FUNCTION LowerBound(Cat: Category; Var SS: TSubSite): Double; {lower bound for category in M above MTL}
     FUNCTION UpperBound(Cat: Category; Var SS: TSubSite): Double; {lower bound for category in M above MTL}

     {End INITX.INC}

     {SUMFRAC.INC:  Procedures that handle writing of output}
     PROCEDURE CSV_file_save(datary:   OutputArray; rowlabel: str_vector1;
                        collabel: str_vector; numrow:   integer; numcol:   integer; fnm: String);
     Procedure Summarize(VAR Map: MSArray; VAR Tropical: Boolean; VAR Summary: OutputArray; Year: Integer);
     Function AggName(CatNum: Integer): String;
     Function Aggregated(i,CatNum: Integer): Double;

     {PROCEDURES THAT HANDLE MEMORY MAP HANDLING CONTAINED IN MSVARRAY.INC}
     Procedure RetA(Row, Col: WORD; Var Cell: CompressedCell);
     Procedure SetA(Row, Col: WORD; Var Cell: CompressedCell);
     Procedure MakeMem(Cell: CompressedCell);
     PROCEDURE RetMem(VAR F: MSArray; Row, Col: WORD; Var Cell: CompressedCell);
     Procedure SetMem(VAR F: MSArray; Row, Col: WORD; Var Cell: CompressedCell);
     Procedure DisposeMem;
     Procedure LoadFile(VAR H: THandle; filen: String);
     Function MakeFile(VAR H: THandle; Rows, Cols: WORD; filen: String; Cell: CompressedCell): Boolean;
     Procedure RetFile(VAR H: THandle; Row, Col: Int64; VAR Cell: CompressedCell);
     Procedure SetFile(Row, Col: Int64; VAR Cell: CompressedCell);
     Procedure DisposeFile(VAR H: THandle; Delete_file: BOOLEAN; FileN: String);
     FUNCTION GetBit(N: BYTE; Row, Col: INTEGER): BOOLEAN;
     Procedure SetBit(N: BYTE; Row, Col: INTEGER; SetTrue: BOOLEAN);
     Function RetCellMatrix(Row, Col: Int64; Var Cell: CompressedCell):Boolean;  {returns optimized cell based on matrix, not map}
     {END OF MSVARRAY.INC}

     {PROCEDURES THAT HANDLE ESTUARY PROCESSING CONTAINED IN FRACFRESH.INC}
     Function  FWInfluenced(Row,Col,FWNum:Integer): Boolean; overload;
     Function  FWInfluenced(Row,Col:Integer): Boolean; overload;
     Function  RiverKM(PR,PC,FWNum: Integer; Var D2C: Double): Double;
     Function CalcSalinity(RunChngwater, TimeZero: Boolean):Boolean;  {Set tidal salinities for all cells}
     Function AggrSalinityStats: Boolean;
     Function FreshWaterHeight(ThisCell: PCompressedCell; FR,FC: Integer;
                               Var FWSal, SWSal: Double; Var Salh: Double;
                               TideInt: Integer; Var SegSalin:Double): Double;
     Function CalculateFlowGeometry(Yr: Integer): Boolean;
     Procedure Smooth_XS_Salinity;
     Procedure ClearFlowGeometry;
     Function CalculateEucDistances: Boolean;
     Function CalculateProbSAV: Boolean;     
     {END OF FRACFRESH.INC}

   End;


function SetFilePointerEx(hFile: THandle; lDistanceToMove: Int64;
  lpNewFilePointer: Pointer; dwMoveMethod: DWORD): DWORD; stdcall;
{$EXTERNALSYM SetFilePointerEx}


VAR

//  SS: TSLAMM_Simulation;
  EachRow,
  EachCol : Integer;


implementation

Uses DrawGrid, FFForm, ElevHistForm, Variants, System.UITypes;

VAR
{  NumRows: Integer; {Number of time-steps to be run}
  Hectares : Double;

  FName: File_Name_Type;


CONST
  WaveLabel: ARRAY[WaveDirection] OF STRING[5] = ('east','west','south','north');
  MenuSelect : ARRAY[1..8] OF STRING[10]
             = ('UnConsol.','Consolid.','UnProt','Bulkhead',
                'Levee','LowDike','HighDike','HardDike');


{$I MSVArray.Inc  }
{$I FracFresh.Inc }
{$I Initx.Inc     }
{$I Transfer.Inc  }
{$I SumFrac.Inc   }
{$I ReadWrite.Inc }
{$I Fractal.Inc   }

{-----------------------------------------------------------------------------------}

(**********************************)
(*   get data and initialize run  *)
(**********************************)

Procedure InitCell(C: PCompressedCell);
Var i: Integer;
Begin
  FOR i := 1 to NUM_CAT_COMPRESS DO
    Begin
      C.Minelevs[i] := 999;
      C.Widths  [i] := 0;
      C.Cats    [i] := Blank;
    End;

  C.TanSlope  := 0.0;

  C.ProtDikes := False;
  C.ElevDikes := False;
  C.MaxFetch := -1.0;  {Set in ChngWater prior to each time-step}
  C.DikeIndex := -99;
  C.SalHeightMLLW := -99;
  For i := 1 to NUM_SAL_METRICS do
    C.Sal[i] := -999;

  C.D2Channel := -99;
  C.Uplift := 0;
  C.ErosionLoss := 0;
  C.BTFErosionLoss := 0;

  C.MTLminusNAVD := -9999;
  C.D2MLLW := -9999;
  C.D2MHHW := -9999;
  C.ProbSAV := -9999;

   C.SubSiteIndex := -9999;
// C.MinFWFlowIndex := -9999;

End;


{-----------------------------------------------------------------------------------}

Function TSLAMM_Simulation.MakeFileName: String;
Var FName: String;
    UncertIter: Integer;
BEGIN { MakeFileName }
  // To be optimal, first check whether outputfileN has a directory specified (search for ":"?}
  // If not, we should prepend the directory of the SLAMM6 file to the MakeFileNane (don't touch outputfilen variable)


  If OutputFileN <> '' then FName := OutputFileN
                       else FName := ElevFileN + 'OUT';

  If UncAnalysis  then With UncertSetup do
    Begin
      if UncSens_Iter=0 then UncertIter := 0
                        else UncertIter := GIS_Start_Num + UncSens_Iter -1;

      FName := ExtractFilePath(CSV_Path)+IntToStr(UncertIter)+'_'+ExtractFileName(CSV_Path);
    End;

  If SensAnalysis then With UncertSetup do FName := xls_path;

  FName := ChangeFileExt(FName,'');
  If Year>0 then FName := FName + ', '+IntToStr(Year)+','
            else FName := FName + ', Initial Condition ';

  If (Not UncAnalysis) and (Not SensAnalysis) and (Year>0) then
    Begin
      If Running_Fixed then FName := FName + ' '+ LabelFixed[FixedScen]+' '
        Else If Running_Custom then FName := FName + ' '+ FloatToStrF(CustomSLR ,ffgeneral,4,2)+ ' meters '
        Else FName := FName + ' '+ LabelIPCC[IPCCSLRate]+' '+LabelIPCCEst[IPCCSLEst]+' ';

      IF ProtectAll THEN
        FName := FName + 'Protect All Dry Land'
      ELSE IF ProtectDeveloped THEN
        FName := FName + 'Protect Developed Dry Land'
      ELSE
        FName := FName + '';

      IF Not IncludeDikes then FName := FName +' No Dikes';
    End;

  If (SensAnalysis) and (UncertSetup.UncSens_Iter>0) {i.e. not deterministic} then
    Begin
      FName := FName + 'Sens'+IntToStr(UncertSetup.UncSens_Iter) + '_'+IntToStr(UncertSetup.PctToVary);
      If UncertSetup.SensPlusLoop then FName := FName + '_Pos'
                                  else FName := FName + '_Neg';
    End;


  MakeFileName := FName;
END;


{ MakeFileName }


{-----------------------------------------------------------------------------------}

Procedure TSLAMM_Simulation.SaveRaster(FileN: String; rasterType: Integer);  {raster Type 1 = Dikes, Raster Type 2 = NWI, Type 3 = Elevations, else Raster-subsites  }
Var File2                   : TextFile;
    CatInt, RowLoop,ColLoop : Integer;
    Current_Cell            : CompressedCell;
    Elev                    : Double;
    TOD                     : TOpenDialog;
    SSIndex : integer;
Begin


  If FileN = ''
    then Begin
            TOD := TOpenDialog.Create(nil);
            TOD.Title:='Select File to Save Raster Output';
            TOD.Filter := '';
            If Not TOD.Execute then
              Begin
                MessageDlg('File Not Saved.',mterror,[mbok],0);
                TOD.Free;
                Exit;
              End;
            FileN := TOD.FileName;
            TOD.Free;
         End;

  If FileExists(FileN) then MessageDlg('Please Backup Original File Before Continuing',mtwarning,[mbok],0);

  AssignFile(File2,FileN);
  Rewrite(File2);
  Writeln(File2,'ncols ',Site.Cols);
  Writeln(File2,'nrows ',Site.Rows);
  Writeln(File2,'xllcorner ',Site.LLXCorner);
  Writeln(File2,'yllcorner ',Site.LLYCorner);
  Writeln(File2,'cellsize ',FloatToStr(Site.Scale));
  Writeln(File2,'nodata_value -99');

  ProgForm.Show;
  ProgForm.ProgressLabel.Caption := 'Saving Raster File';

  {Write EACH CELL}
     For RowLoop := 0 {(Site.Rows Div 2) -1 } to Site.Rows-1 do  //commented out slices in half
      Begin
       For ColLoop:= 0 to Site.Cols-1 do
         Begin
            If ColLoop=1 then If Not ProgForm.Update2Gages(Trunc((RowLoop)/Site.Rows*(100))) then Break;  {break out of loop and close file}

            RetA( RowLoop, ColLoop, Current_Cell);
            case rasterType of
              1:
                Begin
                  if ClassicDike then
                    Begin
                      If CellWidth(@Current_Cell,Blank) > 0  { This code writes dikes }
                        then CatInt := -99
                        else If Current_Cell.ProtDikes
                          then CatInt := 1
                          else CatInt := 0;
                      Write(File2,IntToStr(CatInt),' ');
                    End
                   else
                    Begin
                      If CellWidth(@Current_Cell,Blank) > 0  { This code writes dikes }
                        then  Write(File2,'-99 ')
                        else
                          Begin  //not blank
                            If not Current_Cell.ProtDikes
                              then Write(File2,'0 ')
                              else If Current_Cell.ElevDikes
                                        then Write(File2,floattostrf(GetMinElev(@Current_Cell),ffgeneral,6,6),' ')
                                        else Write(File2,'-5 ');
                          End;
                    End;
                End;
              2:
                Begin
                   CatInt := ORD(GetCellCat(@Current_Cell))+1;
                      If (CatInt <1) or (CatInt > ORD(LastCategory)+1) then
                        Begin
                          CatInt := -99;  {Errant NWI Code being written so write No Data for now}
                        End;
                   Write(File2,CatInt,' ');  {THIS CODE WRITES NWI }
                end;
              3:                                                                         // 2/11/09  overwrite elevation file
                Begin
                  Elev := GetMinElev(@Current_Cell);
                  Write(File2,floattostrf(Elev,ffgeneral,6,6),' ');
                end;
              4:
                begin
                 SSIndex := Site.GetSubSiteNum(ColLoop,RowLoop);
                 Write(File2,inttostr(SSIndex),' ');
                end;


  {            else
                Write(File2,Current_Cell.RastSubS,' ');  }
            end; {case}
          End; {ColLoop}
         Writeln(File2);
        End;
     CloseFile(File2);
  ProgForm.Cleanup;

End;

{-----------------------------------------------------------------------------------}

Function TSLAMM_Simulation.Calc_Connectivity(testR,testC: Integer): Boolean;
var  Cl: CompressedCell;
     Stack, Stack2: TPointStack;

   Function RelevantClass:Boolean;  // a class relevant to the connectivity module, that is to say, if below the salt elevation assumed to be flooded
   var fresh_width: double;
   begin
     fresh_width:= (CellWidth(@cl,DevDryLand) + CellWidth(@cl,UndDryLand) + CellWidth(@cl,InlandFreshMarsh)
           + CellWidth(@cl,Swamp)+CellWidth(@cl,CypressSwamp)+CellWidth(@cl,InlandOpenWater));   // classes that could be protected by dikes and impoundments

     result:= (fresh_width > 0);
   end;

   Function BELOWSALTBOUND(Cl: PCompressedCell; EC,ER: integer): Boolean;
    var subsite: TSubSite;
    begin
       SubSite := Site.GetSubSite(EC,ER,Cl);
       result:= getMinElev(Cl) < SubSite.SaltElev;
    end;

   PROCEDURE CONNECT_CHECK(fx,fy: Integer);    // using a stack fill, check an entire area for connectivity

    function MP(ax,ay:integer): TPoint;
    begin
      result.x:= ax;
      result.y:= ay;
    end;

     Var TP: TPoint;     //for my b.h.
         fillResult: byte;
     begin
       FillResult := 4; {checked & unconnected}


       stack.push(MP(fx,fy));
       stack2.push(MP(fx,fy));

       repeat
         TP:= stack.pop;      // TPoint
         if tp.x > -99 then
            begin
               FX:= TP.x;
               FY:= TP.y;
               IF (FY<0) or (FX<0) or (FX>=Site.Cols) or (FY>=Site.Rows) then continue;  // the fill algorithm has left the map, no connectivity here
               If not (Connect_Array[(Site.Cols*TP.Y)+TP.X] in [1,5,8]) then continue;   // the fill algorithm has found a high area, a diked area, or an area marked as unconnected, no pathway here
               RetA(FY,FX,Cl);
               If OPENWATER(@CL) then Begin
                                       FillResult := 3; {checked & Connected.  We found water so after we've located the whole group we'll mark these as connected}
                                       Connect_Array[(Site.Cols*TP.Y)+TP.X] := 8;  // mark this as open water
                                       Continue;
                                     End;
               //Diked areas not relevant to connectivity algorithm
               if ClassicDike and CL.ProtDikes then Continue;    //ClassicDike case
               if (not ClassicDike) and (CL.ProtDikes) and (not CL.ElevDikes) then Continue; //Not ClassicDike case

               If not BELOWSALTBOUND(@CL, FX, FY) then
                 Begin
                   Connect_Array[(Site.Cols*TP.Y)+TP.X] := 2;
                   Continue;   {Exit if you left the low lying area}
                 End;
               if GetCellCat(@Cl) = blank then
                 Begin
                   Connect_Array[(Site.Cols*TP.Y)+TP.X] := 7;
                   continue;
                 End;

               Stack2.Push(MP(fx,fy));  {Push cell on stack to be marked later}
               Connect_Array[(Site.Cols*TP.Y)+TP.X] := 6; {in process of being checked}

               Stack.Push(MP(FX+1,FY));
               Stack.Push(MP(FX-1,FY));
               Stack.Push(MP(FX,FY+1));
               Stack.Push(MP(FX,FY-1));
               if ClassicDike then
                begin
                  Stack.Push(MP(FX+1,FY+1));
                  Stack.Push(MP(FX-1,FY+1));
                  Stack.Push(MP(FX+1,FY-1));
                  Stack.Push(MP(FX-1,FY-1));  {eight sided search}
                end;
            end;

       until TP.x = -99;

       Repeat
          TP := Stack2.Pop;
          If tp.x > -99 then
           if Connect_Array[(Site.Cols*TP.Y)+TP.X] <> 8 then        {Keep tidal water marked as such for mapping}
              Connect_Array[(Site.Cols*TP.Y)+TP.X] := FillResult;   {utilizing a one dimensional dynamic array for two dimensional data}
       Until TP.X = -99;
    End;

var er,ec: integer;
    cellcat: Category;
begin {Calc_Connectivity}
   Result := True;
   stack  := TPointStack.init;
   stack2 := TPointStack.init;      // result stack

  if Connect_Array = nil then
    setlength(Connect_Array,site.rows*site.cols);
  For ER:=0 to Site.Rows-1 do
    For EC:=0 to Site.Cols-1 do
      Connect_Array[(Site.Cols*ER)+EC]:= 1;     // Initialize presearch, every cell marked as "unchecked"

  ProgForm.ProgressLabel.Caption:='Checking Connectivity :';
  ProgForm.Show;

  For ER:=0 to Site.Rows-1 do
    For EC:=0 to Site.Cols-1 do
    begin
     If EC=0 then Result := ProgForm.Update2Gages(Trunc(ER/Site.Rows*100));
     If Not Result then Begin stack.destroy; stack2.destroy; Exit; End;    // user exit

     if Connect_Array[(Site.Cols*ER)+EC]= 1 then
     begin
       RetA(ER,EC,Cl);
       cellcat := GetCellCat(@Cl);

       if ClassicDike and CL.ProtDikes then Connect_Array[(Site.Cols*ER)+EC]:= 7       //Diked under "Classic" Dike model, don't check.
        else if (not ClassicDike) and (CL.ProtDikes) and (not CL.ElevDikes) then Connect_Array[(Site.Cols*ER)+EC]:= 7 //Not ClassicDike but protected by classic dike designation.
          else if cellcat = blank then Connect_Array[(Site.Cols*ER)+EC]:= 7            // blank, don't check
            else if not BELOWSALTBOUND(@CL, EC, ER) then Connect_Array[(Site.Cols*ER)+EC]:= 2  // above salt bound, no check necessary
              else if not relevantclass then Begin
                                               If OpenWater(@Cl) then Connect_Array[(Site.Cols*ER)+EC]:= 8  // 5/21/2012, show tidal water as navy on maps
                                                                 else Connect_Array[(Site.Cols*ER)+EC]:= 5  // this class is not one we need to check connectivity for e.g. salt marsh or tidal flat
                                             End
                else Connect_Check(EC,ER);
     end;
    end;

  stack.destroy;
  stack2.destroy;

  // ProgForm.Hide;
end;   {Calc_Connectivity}
{-----------------------------------------------------------------------------------}

Function TSLAMM_Simulation.Save_GIS_Files: Boolean;

Type PLoopType =(TWater,TDryland,TSwamp,TFreshWet,
                 TSaltmarsh, TTidalFlat,TBeach);

Var File2                   : TextFile;
    RootName,DescString     : String;
    RowLoop,ColLoop         : Integer;
    Current_Cell            : CompressedCell;
    CatInt                   : Integer;
    WritingElevs, WritingSalin  : Boolean;
    WriteLoop                : Integer;
    SRow, ERow, SCol, ECol   : Integer;
    LLXOut, LLYOut , MinElev : Double;
    ElevOut, SlopeAdjustment : Double;

BEGIN
 Result := True;

 WritingElevs := SaveElevsGIS and not (UncAnalysis or SensAnalysis);
 WritingSalin := SaveSalinityGIS and not (UncAnalysis or SensAnalysis);

 For WriteLoop := 1 to 3 do
    BEGIN
     If (WriteLoop = 2) and not WritingElevs then Continue;
     If (WriteLoop = 3) and not WritingSalin then Continue;

     RootName := MakeFileName;

     ProgForm.ProgressLabel.Caption:='Writing GIS Output :';
     ProgForm.Show;

     DescString := '';
     If (WriteLoop=2) then DescString := '_ELEVS';
     If (WriteLoop=3) then DescString := '_SALINITY_30D_HIGH';// '_SALINITY_30D_HIGH';

     SRow := 0;
     ERow := Site.Rows-1;
     SCol := 0;
     ECol := Site.Cols-1;
     LLXOut := Site.LLXCorner;
     LLYOut := Site.LLYCorner;

    (*  OUTPUT DOMINANT CATEGORY FOR READING INTO GIS  *)
     AssignFile(File2,RootName+DescString +'_GIS.ASC'); // rootname
     Rewrite(File2);
     Writeln(File2,'ncols ',ECol-SCol+1);
     Writeln(File2,'nrows ',ERow-SRow+1);
     Writeln(File2,'xllcorner ',LLXOut);
     Writeln(File2,'yllcorner ',LLYOut);
     Writeln(File2,'cellsize ',FloatToStr(Site.Scale));
     If WriteLoop = 3 then Writeln(File2,'nodata_value -999')
                      else Writeln(File2,'nodata_value -9999');

     {READ EACH CELL}
     For RowLoop := SRow to ERow do
      Begin
        For ColLoop:= SCol to ECol do
         With Current_Cell do
          Begin
            If ColLoop=ECol then
              Begin
                Result := ProgForm.Update2Gages(Trunc((RowLoop-SRow+1)/(ERow-SRow+1)*(100)));
                If Not Result then Begin CloseFile(File2); Exit;  End;  //close file and exit;
              End;

            RetA(RowLoop, ColLoop, Current_Cell);

            Case WriteLoop of
               1: //write SLAMM Codes not Elevations
                  Begin
                    CatInt:=Ord(GetCellCat(@Current_Cell))+1;
                    If CatInt=ORD(Blank)+1 then CatInt:=-9999; {blank}
                    Write(File2,IntToStr(CatInt),' ');
                  End;

               2: // write elevations
                   Begin
                    SlopeAdjustment := (Site.Scale*0.5) * Current_Cell.TanSlope;    {QA 11.10.2005}
                    MinElev := GetMinElev(@Current_Cell);
                    If (MinElev = 999)
                      then Write(File2,'-9999 ')  // no data elevation
                      else
                        Begin
                          ElevOut := GetMinElev(@Current_Cell) + SlopeAdjustment;     {Convert to mean elevation}
                          Write(File2,FloatToStrF(ElevOut,ffgeneral,5,5),' ');
                        End;
                  End;

               3: // write salinity
                   Begin
                    ElevOut := Current_Cell.Sal[4];  //THIS CODE WRITES SALINITY AT 30D Inundation level
                    Write(File2,FloatToStrF(ElevOut,ffgeneral,5,5),' ');
                  End
            End; // case

          End;
        Writeln(File2);
      End;
     CloseFile(File2);
   END;

 // ProgForm.Hide;
END;

{-----------------------------------------------------------------------------------}
PROCEDURE TSLAMM_Simulation.Save_CSV_File;
BEGIN
  If BatchMode then FName := BatchFileN
               else FName := MakeFileName + '.CSV';
  CSV_File_Save(Summary,RowLabel,ColLabel,CountIter+1,ORD(LastCategory)+1,FName);

  UncertSetup.UncSens_Row := CountIter + 1;
END;

{-----------------------------------------------------------------------------------}

Procedure TSLAMM_Simulation.Count_MMEntries;
Begin
  MakeDataFile(True,'','');
End;

{-----------------------------------------------------------------------------------}

Procedure TSLAMM_Simulation.CheckForAnnualMaps;
Var NewDikFName, DikFExt: String;
    NewSalFName, SalFExt: String;
    i: integer;
    buttonSelected : Integer;
begin
  if IncludeDikes=True then
   begin
    NewDikFName := DikFileN;
    DikFExt := ExtractFileExt(DikFileN);
    Delete(NewDikFName,Length(NewDikFName)-Length(DikFExt)+1,Length(DikFExt));    // delete to before the period in the extension eg "test.asc" becomes "test"
    NewDikFName := NewDikFName + IntToStr(Year) + DikFExt;                        // add the year and extension so "test.asc" becomes "test2000.asc"
    if not FileExists(NewDikFName) then NewDikFName:='';
   end;

  if Pos('.xls',lowercase(SalFileN))=0 then
    Begin
      NewSalFName := SalFileN;
      SalFExt := ExtractFileExt(SalFileN);
      Delete(NewSalFName,Length(NewSalFName)-Length(SalFExt)+1,Length(SalFExt));    // delete to before the period in the extension eg "test.asc" becomes "test"
      NewSalFName := NewSalFName + IntToStr(Year) + SalFExt;                        // add the year and extension so "test.asc" becomes "test2000.asc"
      if not FileExists(NewSalFName) then
        begin
          NewSalFName:='';
          if trim(SalFileN)<>'' then
            begin
              buttonSelected := MessageDlg('Warning! Are you sure you want to run this SLAMM simulation with a constant salinity raster?',mtWarning, [mbYes, mbNO], 0);
              if buttonSelected = mrNO then ProgForm.SimHalted := True;
            end;
        end;
    End;

  if (NewDikFName<>'') or (NewSalFName<>'') then
    begin
      MakeDataFile(False,NewDikFName,NewSalFName);
      if NewSalFName<>'' then
        for i := 0 to NumFWFlows-1 do  // Retention Time Must Be Recalculated.
          FWFLows[i].RetentionInitialized := False;
    end

End;




{-----------------------------------------------------------------------------------}

Function TSLAMM_Simulation.MakeDataFile(CountOnly: Boolean; YearDikeFName, YearSalFName: String): Boolean;
Var DikOnly: Boolean;
    SalOnly: Boolean;
    UpliftFile, NWIFile,SLPFile,ElevFile,DikFile,SalFile,IMPFile,ROSFile,VDFile: Textfile;
    NoDataDik, NoDataIMP,NoDataLU, NoDataNWI,NoDataNWIDate,NoDataSal, NoDataROS: Integer;
    NoDataElev, NoDataSlope, NoDataUplift, NoDataVD : Double;
    ElevFExists, SLPFExists, NWIFExists, DIKFExists,ROSFExists,IMPFExists,SalFExists, UpliftFExists,VDFExists: Boolean;

    i,ER,EC: Integer;
    AddToMap: Boolean;
    Neg_Num: Integer;
    ReadStr: String;
    LandMovement: Double;
    PCT_IMP, NWI_Number, ROS_Number: Integer;
    Sal_Number: single;
    NWI_Number2 : Integer;
    Elev_Number, Slope_Number, Dik_Number: double;  {elevation in meters, slope in degrees}
    TotalWidth, SlopeAdjustment: double;
    ReadCat: Category;
    Hist_Adj, DEM_to_NWI_m: Double; {correction in meters}
    ncells: Integer;
    VDNumber, MTL_Correction: Double;
    NRosArray: integer;

Var ReadCell, TemplateCell: CompressedCell;
    NWI_DateSum,NWI_DateCount: Int64;
    SubSite: TSubSite;
    NWI_Date: Integer;
    LF: Integer;

    {---------------------------------------------------------------------------------------------------}
    Function GetNextNumber(var RF: Textfile):double;
    Var S: ANSIString;
        C: ANSIChar;
        FT: ANSIString;
    Begin
      S:='';
      Repeat
        Read(RF,C);
        If (C<>' ') and (ord(C)>32) then
            S := S + C;
      Until ((S<>'') and (C=' ')) or eof(rf);
      Try
        GetNextNumber:=StrToFloat(S);
      Except
        FT := 'raster';
        If (TTextRec(RF).Handle = TTextRec(UpliftFile).Handle) then FT := 'Uplift';
        If (TTextRec(RF).Handle = TTextRec(NWIFile).Handle) then FT := 'SLAMM Category';
        If (TTextRec(RF).Handle = TTextRec(SLPFile).Handle) then FT := 'Slope';
        If (TTextRec(RF).Handle = TTextRec(ElevFile).Handle) then FT := 'DEM';
        If (TTextRec(RF).Handle = TTextRec(DikFile).Handle) then FT := 'Dike';
        If (TTextRec(RF).Handle = TTextRec(ROSFile).Handle) then FT := 'Raster Output Site';
        If (TTextRec(RF).Handle = TTextRec(IMPFile).Handle) then FT := 'Impervious';
        If (TTextRec(RF).Handle = TTextRec(VDFile).Handle) then FT := 'VDatum';
        If (TTextRec(RF).Handle = TTextRec(SalFile).Handle) then FT := 'Salinity';

        Raise ESLAMMError.Create('Error Reading '+FT+ ' File.  Row:'+IntToStr(ER+1)+
              '; Col:'+IntToStr(EC+1)+'. "'+ S+'" is not a valid number.');
      End;
    End;
    {---------------------------------------------------------------------------------------------------}
    Procedure PrepareFileForReading(Var FN: String; Var FExists: Boolean; Var FNoDat:Double; Var TFile:Text); overload;
    Begin
      FNoDat := -9999;
      FExists := (FN <> '');
      If FExists then if not FileExists(FN) then
        Begin
          ProgForm.Hide;
          Raise ESLAMMError.Create('File Setup Error, Cannot find file "'+FN+'"');
        End;

      If FExists then
        Begin
          Assign(TFile,FN);
          Reset(TFile);

          Repeat ReadLn(TFile,ReadStr)       {Parse through header lines if any}
          Until eof(TFile) or (Pos('nodata_value',LowerCase(ReadStr))>0);

          If Pos('nodata_value',LowerCase(ReadStr))=0 then
          Raise ESlammError.Create('Data File "'+FileN+'" Does Not Include "nodata_value" identifier');

          Delete(ReadStr,Pos('nodata_value',LowerCase(ReadStr)),12);
          FNoDat:=StrToFloat(ReadStr);
        End;
     End;
    {---------------------------------------------------------------------------------------------------}
     Procedure PrepareFileForReading(Var FN: String; Var FExists: Boolean; Var FNoDat:Integer; Var TFile:Text);  overload;
     Var DubNoDat: Double;
     Begin
        PrepareFileForReading(FN,FExists,DubNoDat,TFile);
        FNoDat := Trunc(DubNoDat);
     End;
    {---------------------------------------------------------------------------------------------------}
    Procedure ReadDikeOrSalinityOnly;
          Begin
            // Read the cell attributes
            RetA(ER,EC,ReadCell);

            // Update dike value
            if DikOnly then
              begin
                ReadCell.ProtDikes := (Dik_Number <> NoDataDik) and ((DIK_Number >0) or (DIK_Number=-5));
                if (not ClassicDike) then
                  ReadCell.ElevDikes := (DIK_Number <> NoDataDik) and (DIK_Number > 0);
              end;

            // Update salinity value
            if SalOnly then
              ReadCell.Sal[1] := Sal_Number;

            // Set cell attributes
            SetA(ER,EC,ReadCell);
          End;
    {---------------------------------------------------------------------------------------------------}

Var FlatArrIndex: LongInt;

Begin  {MakeDataFile}
  Result := True;
  DikOnly := YearDikeFName <> '';
  SalOnly := YearSalFName <> '';

  if CountOnly and (optimizelevel = 0) then
    begin
      NumMMEntries := Site.rows * Site.cols;
      exit;
    end;

  TotalWidth := Site.Scale;
  Site.MaxROS := 0;  {Default to zero raster output subsites}
  Site.ROSBounds := nil; {empty dynamic array}
  MaxMTL := -9999;
  MinMTL := 9999;

  If BatchMode then ProgForm.Caption := 'Simulating '+Site.GlobalSite.Description;
  If (Not DikOnly) and (Not SalOnly) then ProgForm.YearLabel.Visible:=False;
  ProgForm.SLRLabel.Visible:=False;
  ProgForm.ProtectionLabel.Visible:=False;
  ProgForm.Show;
  ProgForm.Update;

  InitCell(@TemplateCell);

  If (Not CountOnly) and (Not DikOnly) and (Not SalOnly) then
    Begin
      S60FileN:= 'SLAMM6_Disk';
      If SaveMapsToDisk and MakeNewDiskMap then If Not MakeFile(MapHandle,Site.Rows,Site.Cols,S60FileN,TemplateCell) then Exit;
      If Not SaveMapsToDisk
          Then
             Try
               MakeMem(TemplateCell);
             Except
               If Map<>nil then
                 Begin
                   LF := Length(Map);
                   For i := 0 to LF-1 do
                     Dispose(Map[i]);
                   Map := nil;
                 End;
               Map := nil;
(*               If MessageDlg('Error creating map in memory, write map to disk instead?',mtconfirmation,[mbyes,mbno],0) = mryes
                 then Begin
                        SaveMapsToDisk := True;
                        MakeFile(MapHandle,Site.Rows,Site.Cols,S60FileN,TemplateCell);
                      End
                 else *)
                      Begin
                        ProgForm.Cleanup;
                        Raise ESLAMMError.Create('Error creating map in memory, map exceeds size of memory in system.');
                      End;
             End;
      ProgForm.Show;
      ProgForm.Update;

    End;

  ProgForm.ProgressLabel.Caption:='Reading Site Characteristics';

  If CountOnly then ProgForm.ProgressLabel.Caption:='Counting Cells to Track';

  // Prepare the (annual) dike files to read if they exist
  If DikOnly then
    begin
      ProgForm.YearLabel.Caption:=IntToStr(Year);
      ProgForm.ProgressLabel.Caption:='Reading Dike File '+ExtractFileName(YearDikeFName);
      PrepareFileForReading(YearDikeFName,DikFExists,NoDataDik,DikFile);     //open annual dike file
    end
  else PrepareFileForReading(DikFileN,DikFExists,NoDataDik,DikFile);       // 3.11.2010 always load dike file for pre-processor under "no-dikes" conditions

  SalFExists := False;
  if Pos('.xls',lowercase(SalFileN))=0 then
    Begin
      //Prepare the (annual) salinity file(s) to read if they exist
      if SalOnly then
        begin
          ProgForm.YearLabel.Caption:=IntToStr(Year);
          ProgForm.ProgressLabel.Caption:='Reading Salinity File '+ExtractFileName(YearSalFName);
          PrepareFileForReading(YearSalFName,SalFExists,NoDataSal,SalFile);   //open annual salinity file
        end
      else PrepareFileForReading(SalFileN,SalFExists,NoDataSal,SalFile);
    End;

  // Prepare files to read when is not only a dike and not salinity map only
  if (not DikOnly) and (not SalOnly) then
     Begin
        PrepareFileForReading(ROSFileN,ROSFExists,NoDataROS,ROSFile);
        If ROSFExists
           then Begin
                  //NRosArray := Site.Rows*Site.Cols;
                  //if (Length(ROSArray) < NRosArray) then SetLength(ROSArray, NRosArray);
                  if (Length(ROSArray) < Site.Rows*Site.Cols) then SetLength(ROSArray, Site.Rows*Site.Cols);
                  For i := 0 to Site.Rows*Site.Cols -1 do
                    ROSArray[i] := 0;
                End
           else ROSArray := nil;
        PrepareFileForReading(IMPFileN,IMPFExists,NoDataIMP,IMPFile);
        PrepareFileForReading(NWIFileN,NWIFExists,NoDataNWI,NWIFile);
          If Not NWIFExists then Raise ESlammError.Create('Must include SLAMM Code (NWI) Data File');
        PrepareFileForReading(ElevFileN,ElevFExists,NoDataElev,ElevFile);
        Site.NoElevData := NoDataElev;
          If Not ElevFExists then Raise ESlammError.Create('Must include Elevation Data File');
        PrepareFileForReading(SLPFileN,SLPFExists,NoDataSlope,SLPFile);
          If Not SLPFExists then Raise ESlammError.Create('Must include Slope Data File');
        PrepareFileForReading(UpliftFileN,UpliftFExists,NoDataUplift,UpliftFile);
        PrepareFileForReading(VDFileN,VDFExists,NoDataVD,VDFile);
     End;

  // If either CountOnly or DikOnly or SalOnly is false
  if not (CountOnly or DikOnly or SalOnly) then
    begin
      if (Length(MapMatrix) < Site.Rows*Site.Cols) then SetLength(MapMatrix, Site.Rows*Site.Cols);
      if USE_DATAELEV and (OptimizeLevel>1) and (Length(DataElev) < Site.Rows*Site.Cols) then SetLength(DataElev, Site.Rows*Site.Cols);
      if (OptimizeLevel>1) and (Length(MaxFetchArr) < Site.Rows*Site.Cols) then SetLength(MaxFetchArr, Site.Rows*Site.Cols);
      if (OptimizeLevel>1) then
        for i := 0 to Site.Rows*Site.Cols -1 do
         MaxFetchArr[i] := 0;
    end;

  NCells:=0;

  for ER:=0 to Site.Rows-1 do
    begin

      // Progress bar
      Result := ProgForm.Update2Gages(Trunc((Er)/Site.Rows*100));
      If not Result then Exit;

      for EC:=0 to Site.Cols-1 do

      begin
        FlatArrIndex := (Site.Cols*ER)+EC;

        InitCell(@ReadCell);
        SubSite := Site.GetSubSite(EC,ER,@ReadCell);

        //Dike data
        DIK_Number:=0;
        If DikFExists then DIK_Number := GetNextNumber(DIKFile);      //Trunc(GetNextNumber(DIKFile))

        //Salinity Data
        Sal_Number :=0;
        if SalFExists then Sal_Number := GetNextNumber(SalFile);

        If DikOnly or SalOnly then          // If only salinity or dike are being updated mid-run
          Begin
            ReadDikeOrSalinityOnly;
            Continue;  // continue to the next cell
          End;

        //Raster Output Sites Data
        if (ROSFileN <> '')
          then
            begin
              ROS_Number := Trunc(GetNextNumber(ROSFile));
              if ROS_Number = NoDataROS then ROS_Number := 0;
            end
        else ROS_Number := 0;

        //NWI Date
        NWI_Date := Subsite.NWI_Photo_Date;

        //Elevation Number
        Elev_Number := GetNextNumber(ElevFile);

        //  Elev_Number := Elev_Number - 3.33;  {3.0, 2.75, 3.5 meter test}

(*        If NWI_Number = ORD(InlandFreshMarsh)+1 then NWI_Number := ORD(TidalFreshMarsh)+1;

        {if (Site.GetSubSite(EC,ER) = Site.SubSites[1])
         and (NWI_number = nodataNWI) or (NWI_number = 24)
          then NWI_number := 17;} // code blanks in subsite 2 as Estuarine Water

        {if (Site.GetSubSite(EC,ER) = Site.SubSites[0]) and (NWI_number = 19)
          then NWI_number := 17;}  // code open ocean in subsite 1 as Estuarine Water

       {if (Site.GetSubSite(EC,ER) = Site.SubSites[1])
          and (NWI_number = nodataNWI) or (NWI_number = 24)
           then NWI_number := 19;}// code blank as open water in subsite 1


        // ECL CODE FOR DRY LAND IN FLORIDA
        //if (elev_number <> nodataelev) and (NWI_Number = nodataNWI) then
        // NWI_Number := 2;

       { if (elev_number = nodataelev) and (NWI_Number = 2) then
          NWI_Number := 24;      }
*)


        //Slope Number
        Slope_Number := GetNextNumber(SlpFile);

        //NWI data
        NWI_Number := Trunc(GetNextNumber(NWIFile));

        if NWI_Number = 0 then NWI_Number:= 24;

   //     If (Elev_Number > -1e-10) and (Elev_Number<1e-10) and (NWI_Number in [17,19])
    //       then Elev_Number := -5;   //fixme allga correction for open ocean


        If ((NWI_Number < 1) or (NWI_Number>26)) and (NWI_Number<>NoDataNWI) then
              Raise ESlammError.Create('SLAMM6 Does not understand NWI Class Number '+IntToStr(NWI_Number)+' Within Data File "'+NWIFileN);

        // Read Impervious file if it exists or set to zero
        if IMPFExists then
          begin
            PCT_IMP := Trunc(GetNextNumber(IMPFile));
            ReadCell.ImpCoeff := PCT_IMP;
            IF (PCT_IMP >=0) and (NWI_Number = 1) or (NWI_Number = 2) then
               If PCT_IMP > {33 } 25 {15} {50} then NWI_Number := 1 else NWI_Number := 2;
          end
        else ReadCell.ImpCoeff := 0; //In case the impervious file is not available/loaded into the project

        //MTL Correction
        MTL_correction:= SubSite.NAVD88MTL_correction;
        If VDFExists then
          begin
            VDNumber:= GetNextNumber(VDFile);
            if VDNumber <> NoDataVD then
              MTL_Correction:= VDNumber;
          end;

        For i := 1 to 2 do
          If RunUncertainty and (UncertSetup.ZUncertMap[i] <> nil) and (UncertSetup.UncSens_Iter > 0) then
            If (elev_Number <> NoDataElev) then
              Elev_Number := Elev_Number + UncertSetup.ZUncertMap[i][FlatArrIndex];

        If (NWI_Number = NoDataNWI) then NWI_Number := ORD(Blank)+1;

{          then Begin
                  SetCellWidth(@ReadCell,Blank,TotalWidth);
                  MapMatrix[FlatArrIndex] := -99;
               End
          else Begin }

         // Dike attribute
         ReadCell.ProtDikes := (DIK_Number <> NoDataDik) and ((DIK_Number > 0) or (DIK_Number=-5));
         if (not ClassicDike) then
          begin
           ReadCell.ElevDikes := (DIK_Number <> NoDataDik) and (DIK_Number > 0);
           if ((NWI_Number in [16..19]) or (NWI_Number=NoDataNWI)) and (ReadCell.ElevDikes) then  //estuarine open water, open ocean, riverine tidal, tidal creek
            NWI_Number := 1;
          end;

         //code open water as inland open water behind dikes (TNC-Walsh)
         //if (Site.GetSubSite(EC,ER) = Site.SubSites[9]) and (NWI_Number in [16..19]) then
         //begin
         //   NWI_number := 15;
          //end;

         // Assign the salinity value
         if SalFExists then ReadCell.Sal[1] := Sal_Number;

         // MTL to NAVD88 correction
         ReadCell.MTLminusNAVD := MTL_correction;
         If MTL_Correction > MaxMTL then MaxMTL := MTL_Correction;
         If MTL_Correction < MinMTL then MinMTL := MTL_Correction;

         // Uplift correction
         If UpliftFExists then ReadCell.Uplift := GetNextNumber(UpliftFile)
                          else ReadCell.Uplift := 0;


         ReadCell.ErosionLoss := 0;
         ReadCell.BTFErosionLoss := 0;

         ReadCat := Category(NWI_Number-1);
         If (elev_number = nodataelev) and LoadBlankIfNoElev then
            Readcat:= blank;

         SetCellWidth(@ReadCell,ReadCat,TotalWidth);

         If (not UpliftFExists) or (ReadCell.Uplift = NoDataUplift) then  //no cell by cell uplift available
           with Subsite do
             Begin
               Hist_Adj := Historic_trend - 1.7; { mm/yr global historic trend, subtracted from the local historic trend to remove
                                                  double counting when futurecasting. http://www.grida.no/climate/ipcc_tar/wg1/424.htm }
                                                 { 1.7 based on IPCC 2007a }

               Readcell.Uplift := - Hist_Adj * 0.1;
                  {cm/yr}           {mm/yr}  {cm/mm}
               {projected SLR} {Years of the model run beyond 1990} {Historic local trend - local adjustment if req.}
             End;

         // Set cell slope
         If (Slope_Number <> NoDataSlope) then
           ReadCell.TanSlope:=Tan(DegToRad(Slope_Number));

         // Elevation adjustments
         if (not ClassicDike) and (ReadCell.ElevDikes)  //elevation will be set to dike height.
            then Elev_Number := DIK_Number;

          If (Elev_Number <> NoDataElev) then
             Begin
               if (not ReadCell.ElevDikes) then   //no adjustments in this case
                begin

                  //(1) Refer to MTL
                  Elev_Number := (Elev_Number - MTL_Correction) ; {Set Elevation so that MT = 0.0}
                    {meters}        {meters}       {meters}

                  //(2) Consider uplift for the years between DEM and NWI
                  LandMovement := -(ReadCell.Uplift * 10);
                    {mm/year}          {cm/year  *  mm/cm}

                  With subsite do
                    DEM_to_NWI_m := (NWI_Date - DEM_Date)* (LandMovement) * 0.001 ;
                      {meters}      {       years       }      {mm/yr}    {m / mm}

                  Elev_Number := Elev_Number - DEM_to_NWI_m;  { calculate elevations at NWI photo date }
                    {meters}       {meters}      {meters}

                  //(3) Slope correction
                  SlopeAdjustment := (Site.Scale*0.5) * ReadCell.TanSlope;    {QA 11.10.2005}
                  Elev_Number := Elev_Number - SlopeAdjustment;

                  //(4) Dike elevation
                  //if ReadCell.ElevDikes  then Elev_Number := DIK_NUmber;    //Marco - Levee elevations are provided with respect to MSL
                end;

               // Set elevation
               SetCatElev(@ReadCell,ReadCat,Elev_Number);

               If not CountOnly then
                 IF USE_DATAELEV and (OptimizeLevel>1) then
                   If (Elev_Number) < 0
                     then DataElev[FlatArrIndex] := 0
                     else If (Elev_Number) > 65.5
                       then DataElev[FlatArrIndex] := 65500
                       else DataElev[FlatArrIndex] := Round((Elev_Number)*1000);

             End;  //  If (Elev_Number <> NoDataElev) then

        if ROSFExists then ROSArray[FlatArrIndex] := ROS_Number;

             With Site do  { Assign min and max boundaries for ROS "Raster Output Sites" }
               Begin
                 If MaxROS < ROS_Number then
                   Begin
                     If ROS_Number > Length(ROSBounds) then SetLength(ROSBounds,ROS_Number+2);

                     For i := MaxROS+1 to ROS_Number do
                       Begin
                         ROSBounds[i-1].X1 := 99999;   {min}
                         ROSBounds[i-1].Y1 := 99999;   {min}
                         ROSBounds[i-1].X2 := -99999;  {max}
                         ROSBounds[i-1].Y2 := -99999;  {max}
                       End;
                     MaxROS := ROS_Number;
                   End;

                 If ROS_Number>0 then with ROSBounds[ROS_Number-1] do
                   Begin
                     If X1>EC then X1 := EC;
                     If Y1>ER then Y1 := ER;
                     If X2<EC then X2 := EC;
                     If Y2<ER then Y2 := ER;
                    End;
               End;


        AddToMap := (OptimizeLevel=0) or ((NWI_Number<>NoDataNWI) and (NWI_Number<>ORD(Blank)+1));
        Neg_Num := -99;
        IF (OptimizeLevel>1) and
                ( ( NWI_Number = ORD(OpenOcean)+1 ) or
                  ( NWI_Number = ORD(EstuarineWater)+1 ) or
                  (( NWI_Number = ORD(DevDryLand) +1) and (CatElev(@ReadCell,ReadCat) > ELEV_CUTOFF ))or
                  (( NWI_Number = ORD(UndDryLand) +1) and (CatElev(@ReadCell,ReadCat) > ELEV_CUTOFF )) )  then
          Begin
            Neg_Num := -NWI_Number;
            AddToMap := False;
          End;

        If AddToMap then Inc(NCells);

        If Not CountOnly then
          Begin
            If AddToMap
              then MapMatrix[FlatArrIndex] := NCells-1
              else MapMatrix[FlatArrIndex] := Neg_Num;

             If ((Not SaveMapsToDisk) or (MakeNewDiskMap))
                then SetA(ER,EC,ReadCell);
          End;
      End;
  end; {For ER, For EC loops}

  If DikOnly then
    Begin
      CloseFile(DikFile);
      //Exit;
    End;

  if SalOnly then
     Begin
      CloseFile(SalFile);
      //Exit;
    End;

  if DikOnly or SalOnly then Exit;

  NumMMEntries := NCells;

  InitCell(@BlankCell);
  SetCellWidth(@BlankCell,Blank,Site.Scale);

  InitCell(@OceanCell);
  SetCellWidth(@OceanCell,OpenOcean,Site.Scale);
  SetCatElev(@OceanCell,OpenOcean,- Site.GlobalSite.SaltElev);

  InitCell(@EOWCell);
  SetCellWidth(@EOWCell,EstuarineWater,Site.Scale);
  SetCatElev(@EOWCell,EstuarineWater,- Site.GlobalSite.SaltElev);

  InitCell(@UndDryCell);
  SetCellWidth(@UndDryCell,UndDryLand,Site.Scale);
  SetCatElev(@UndDryCell,UndDryLand,Elev_Cutoff);

  InitCell(@DevDryCell);
  SetCellWidth(@DevDryCell,DevDryLand,Site.Scale);
  SetCatElev(@DevDryCell,DevDryLand,Elev_Cutoff);

  // ProgForm.Hide;

  CloseFile(NWIFile);

  CloseFile(SLPFile);
  CloseFile(ElevFile);

  If DikFExists then CloseFile(DikFile);
  if SalFExists then Close(SalFile);
  If IMPFExists then CloseFile(IMPFile);
  If ROSFExists then CloseFile(ROSFile);


  DisposeFile(MapHandle,False,'');
End;

{-----------------------------------------------------------------------------------}

Function TSLAMM_Simulation.CopyMapsToMSWord : Boolean;

      {-------------------------------------------------------------------------}
      Procedure CopyCurrentMap(GF: TGridForm; Desc: String; Attempt: Integer);
      var
        Descrip2 : String;
        Range: Variant;
        NumPars: Integer;

              Procedure AddandSelect;
              Begin
                WordDoc.Paragraphs.Add;
                NumPars := WordDoc.Paragraphs.Count;
                Range := WordDoc.Range(
                    WordDoc.Paragraphs.Item(NumPars).Range.Start,
                    WordDoc.Paragraphs.Item(NumPars).Range.End);
              End;

      Begin
         Descrip2 := ExtractFileName(MakeFileName) + Desc;

         Try

         If Maps_to_GIF then GF.SaveMaptoGIF(Descrip2);

         If Not Maps_to_MSWord then Exit;

         GF.CopyMapToClipboard(nil);

         If not WordInitialized then
            Begin
              try
                WordApp := GetActiveOLEObject('Word.Application');
              except
                WordApp := CreateOLEObject('Word.Application');
              end;
              WordApp.Visible := True;
              WordApp.Documents.Add;
              WordDoc := WordApp.Documents.Item(1);
              WordDoc.Paragraphs.Add;
              WordInitialized := True;
            End;

         NumPars := WordDoc.Paragraphs.Count;
         WordDoc.Paragraphs.Add;
         WordDoc.Paragraphs.Add;

         Range := WordDoc.Range(
         WordDoc.Paragraphs.Item(NumPars + 2).Range.Start,
         WordDoc.Paragraphs.Item(NumPars + 2).Range.End);
         Range.Paste;

{        WordDoc.Paragraphs.Add; }  {TEST}                  // resize document within word, look into VB for answer on MSWord side
         AddAndSelect;
         Range.Text := Descrip2;
         AddAndSelect;
         Range.Text := DateTimeToStr(Now()) + '  ~';

         Except
           If Attempt<4
             then Begin
                    SysUtils.Sleep(5000);
                    CopyCurrentMap(GF,Desc,Attempt+1);
                  End
             else If MessageDlg('Runtime Error copying map to Microsoft Word for year '+IntToStr(Year)+
                                '.  Retry writing map to Word?',mterror,[mbyes,mbno],0)= mryes
                     then
                       Begin
                         If Attempt > 4 then
                           Begin
                             If MessageDlg('Cannot contact old word document in multiple attempts.  Open New Word document?',mterror,[mbyes,mbno],0)= mryes
                               then WordInitialized := False;
                           End;
                         CopyCurrentMap(GF,Desc,Attempt+1);
                       End;
         End;
      End;
      {-------------------------------------------------------------------------}

Var i: Integer;
    gridform2: TGridForm;

      Procedure PrepareGridForm2(X1,X2,Y1,Y2: Integer; Desc: String; ROS: boolean);
      Var j, jmax: Integer;
          DescLabel: String;
      Begin
           With GridForm2 do
             Begin
               DrawingRect := False;
               RunPanel.Top := 0;
               RunPanel.Visible := True;
               //ShowRunPanel;
               ZoomBox.Visible := False;
               ToolBox1.ItemIndex := 1;

               MapNR := ABS(Y1-Y2)+1;
               MapNC := ABS(X1-X2)+1;
               GridScale := Site.Scale;
               SS := Self;
               ThisSite := Site;
               UserStop:=False;
               GridForm2.SlpFileN := SlpFileN;
               SaveBtStream := False;
               Result := DrawEntireMap(i,True,ROS);
               If Not Result then Exit;
               CopyCurrentMap(GridForm2,' '+Desc,1);

{               GraphBox.Items =
                 0 SLAMM Map
                 1 Elevations
                 2 Salinity MLLW
                 3 Salinity MTL
                 4 Salinity MHHW
                 5 Salinity 30D hi
                 6 Accretion
                 7 Subsidence
                 8 MTL NAVD88
                 9 Marsh Erosion
                 10 Beach Erosion
                 11 Simplified Categ
                 12 Probability SAV }

               If ExtraMaps then {salinity, accretion maps}
                 begin
                  // Salinity
                  jmax := 5;
                  if trim(SalFileN)<>'' then jmax := 2;
                  For j := 2 to jmax do
                    If (NumFWFlows > 0) or ((j<2) or (j>5)) then
                      Begin
                        GridForm2.GraphBox.ItemIndex := j;
                        Result := DrawEntireMap(i,True,ROS);
                        If Not Result then Exit;
                        DescLabel := GridForm.GraphBox.Items[j];
                        if j=2 then DescLabel := 'Salinity';
                        CopyCurrentMap(GridForm2,DescLabel+' '+Desc,1);
                      End;

                  //Accretion
                  GridForm2.GraphBox.ItemIndex := 6;
                  Result := DrawEntireMap(i,True,ROS);
                  If Not Result then Exit;
                  CopyCurrentMap(GridForm2,GridForm2.GraphBox.Items[6]+' '+Desc,1);
                 end;

               If DucksMaps then
                   Begin
                     GridForm2.GraphBox.ItemIndex := 11;
                     Result := DrawEntireMap(i,True,ROS);
                     If Not Result then Exit;
                     CopyCurrentMap(GridForm2,GridForm2.GraphBox.Items[11]+' '+Desc,1);
                   End;

           (*     If i = 3 then with GridForm2 do   //fixme freeport only
                   Begin
                      DontDraw := True;
                      ShowDikes.Checked := True;
                      ConnectBox1.Checked := True;
                      DontDraw := False;
                      ShowDikesClick(ConnectBox1);
                      CopyCurrentMap(GridForm2,'Connect '+Desc,1);
                      DontDraw := True;
                      ShowDikes.Checked := False;
                      ConnectBox1.Checked := False;
                      DontDraw := False;
                      progform.visible := true;
                      GridForm.ShowRunPanel;
                   End; *)


             End;

      End;

Var k, kmax: Integer;
    DescLabel: String;
Begin       // PROCEDURE CopyMapsToMSWord
   Result := True;
   GridForm.SaveBtStream := False;

(*If  UncAnalysis then    //code would limit saving of files to word/GIF to GIS year setup
   Begin
     If Not UncertSetup.SaveUncGIS then Exit;
     If Not UncertSetup.SaveALLGIS then
       if not IntInString(Year,UncertSetup.GISYears) then Exit;
   End; *)

  If GridForm.GraphBox.ItemIndex <> 0 then
    Begin
      GridForm.GraphBox.ItemIndex := 0;
      Result := GridForm.DrawEntireMap(0,True,False);
      If Not Result then Exit;
      GridForm.RunPanel.Visible := True;
    End;

    CopyCurrentMap(GridForm,'',1);
    If ExtraMaps then
      begin
        // Salinity
        kmax := 5;
        if trim(SalFileN)<>'' then kmax := 2;

        For k := 2 to kmax do
          If (NumFWFlows > 0) or ((k<2) or (k>5)) then
          Begin
            GridForm.GraphBox.ItemIndex := k;
            Result := GridForm.DrawEntireMap(0,True,False);
            If Not Result then Exit;
            GridForm.RunPanel.Visible := True;
            DescLabel := GridForm.GraphBox.Items[k];
            if k=2 then DescLabel := 'Salinity';
            CopyCurrentMap(GridForm,DescLabel,1);
          End;

        // Accretion
        GridForm.GraphBox.ItemIndex := 6;
        Result := GridForm.DrawEntireMap(0,True,False);
        If Not Result then Exit;
        GridForm.RunPanel.Visible := True;
        CopyCurrentMap(GridForm,GridForm.GraphBox.Items[6],1);
      end;


    If DucksMaps then
      Begin
        GridForm.GraphBox.ItemIndex := 11;
        Result := GridForm.DrawEntireMap(0,True,False);
        If Not Result then Exit;
        GridForm.RunPanel.Visible := True;
        CopyCurrentMap(GridForm,GridForm.GraphBox.Items[11],1);
      End;

    If DucksMaps or (GridForm.GraphBox.ItemIndex <> 0) then
      Begin
        GridForm.GraphBox.ItemIndex := 0;
        Result := GridForm.DrawEntireMap(0,True,False);
        If Not Result then Exit;
        GridForm.RunPanel.Visible := True;
      End;

   Result := ValidateOutputSites;
   If Not Result then Exit;

   For i := 1 to Site.NOutputSites do  {These are output sites, ROS handled below}
      with Site.OutputSites[i-1] do
         Begin
           If i=1 then Application.CreateForm(TGridForm,GridForm2);
           GridForm2.SaveBtStream := False;
           If UsePolygon then with Poly do PrepareGridForm2(MinCol,MaxCol,MinRow,MaxRow,Description, FALSE)
                         else with Rec do  PrepareGridForm2(X1,X2,Y1,Y2,Description, FALSE);
           If i=Site.NOutputSites then GridForm2.Free;
         End;

   For i := 1 to Site.MaxROS do
    with Site.RosBounds[i-1] do
       Begin
         If i=1 then Application.CreateForm(TGridForm,GridForm2);
         GridForm2.SaveBtStream := False;
         PrepareGridForm2(X1,X2,Y1,Y2,'ROS '+IntToStr(i),TRUE);
         If i=Site.MaxROS then GridForm2.Free;
       End;
End;

{-----------------------------------------------------------------------------------}

Function TSLAMM_Simulation.ValidateOutputSites: Boolean;
var i,j: Integer;
Begin
   Result := True;
   For i := 1 to Site.NOutputSites do
    with Site.OutputSites[i-1] do
     If UsePolygon
      then with Poly do
       Begin
         Result := NumPts>1;
         If Not result then
           Begin
             MessageDlg('Boundary of polygon output site '+inttostr(i)+' not properly defined.',mterror,[mbok],0);
             Exit;
           End;

         For j := 1 to NumPts do with TPoints[j-1] do
           Begin
             if X<0 then X := 0; If X>Site.Cols-1 then X := Site.Cols-1;
             if Y<0 then Y := 0; If Y>Site.Rows-1 then Y := Site.Rows-1;
           End;
       End
      else with Rec do
       Begin
         Result := (X1>0) or (X2>0) or (Y1>0) or (Y2>0);
         If Not result then
           Begin
             MessageDlg('Boundary of output site '+inttostr(i)+' not properly defined.',mterror,[mbok],0);
             Exit;
           End;

         if X1<0 then X1 := 0; If X1>Site.Cols-1 then X1 := Site.Cols-1;
         if X2<0 then X2 := 0; If X2>Site.Cols-1 then X2 := Site.Cols-1;
         if Y1<0 then Y1 := 0; If Y1>Site.Rows-1 then Y1 := Site.Rows-1;
         if Y2<0 then Y2 := 0; If Y2>Site.Rows-1 then Y2 := Site.Rows-1;
       End;

End;

{------------------------------------------------------------------------------}


PROCEDURE TSLAMM_Simulation.SetMapAttributes;
Var i, NPage: integer;
Begin
  GridForm.Hide;
  GridForm.ZoomFactor := InitZoomFactor;

  Connect_Array := nil;

//  ReadSite;

  If SaveMapsToDisk and Not MakeNewDiskMap
    then if not FileExists(S60FileN) then
       Raise ESlammError.Create('Data File "'+S60FileN+'" Does Not Exist!');

  If Not MakeDataFile(False,'','') then Begin ProgForm.Hide; Exit; End;

  Site.InitElevVars;

  If SaveMapsToDisk then
    Begin
      LoadFile(MapHandle,S60FileN);            {Work with Map File}
    End;



  with Gridform do
    begin
      GraphBox.ItemIndex := 0;
      ThisSite := Site;
      SS := Self;

      // Set the initial tab to the first sheet
      PageControl1.ActivePageIndex := 1;
//      PanCheckBox.Checked := True;

      // Set some buttons to unchecked
      ShowDikes.Checked := False;

      // Make execution buttons invisible
      HaltButton.Visible := False;
      NextStepButton.Visible := False;
      StopPausingButton.Visible := False;

      // Make all pages visible
      NPage := PageControl1.PageCount;
{$IFDEF DEBUG}
      for i := 0 to NPage-1 do  //  include last page (WPC utilities).
{$ELSE}
      for i := 0 to NPage-2 do  //  remove last page for release (WPC utilities).
{$ENDIF}
        PageControl1.Pages[i].TabVisible := True;

      //Set Caption of Page zero
      PageControl1.Pages[0].Caption := 'Edit Input or Output Subsites';


      EditAttributes(Site.Rows,Site.Cols, Self);

      If DikesChanged then
        If MessageDlg('Save edits to Dike Raster?  (OVERWRITE)',mtconfirmation,[mbyes,mbno],0) = mryes
          then SaveRaster( DikFileN,1 );

      If NWIChanged then
        If MessageDlg('Save edits to NWI Raster?  (OVERWRITE)',mtconfirmation,[mbyes,mbno],0) = mryes
          then SaveRaster( NWIFileN,2 );

      DikesChanged := False;
      NWIChanged := False;
    end;

  ProgForm.Cleanup;

End;


{------------------------------------------------------------------------------}

PROCEDURE TSLAMM_Simulation.CalibrateSalinity;
Begin
   If (NuMFWFlows < 1) and (trim(SalFileN)='') then
    Begin
      MessageDlg('You must first define Freshwater Flows ("Set Map Attributes") before Calibrating Salinity.',mterror,[mbok],0);
      Exit;
    End;

   If (Init_SalStats=False) and (trim(SalFileN)='') then
    Begin
      MessageDlg('Salinity Statistics must be calculated by running the model to the start of time-zero.',mterror,[mbok],0);
      Exit;
    End;

   If (Init_SalStats=False) and (trim(SalFileN)<>'') then
    Begin
      MessageDlg('Initially, salinity statistics can be run only after loading the map.',mterror,[mbok],0);
      Exit;
    End;
    
  SalinityForm.SS := Self;
  Salinityform.ShowSalinityAnalysis;
  Salinityform.Showmodal;
  ProgForm.Cleanup;

End;

          {==============================================================================}


Procedure TSLAMM_Simulation.Calc_Elev_Stats;
//============================================
// Calculation of elevation statistics
//============================================
Var
    SlopeAdjustment: Double; {elev stat code}
//    ElevFile: TextFile;
    PctInd, Ind,ER,EC, j: Integer;
    CC: Category;
    StatCell: CompressedCell;
    SubSite: TSubSite;
    Values, SortVals: Array[1..8,Category] of Array of Single;
    CE: Double;
    ElevLevel: array[1..8] of Double;
    MinWetElev: Double;

Begin      {Map must be loaded for this calculation}

  // Initialize statistical indicators
  for CC := FirstCategory to Lastcategory do
    begin
      for Ind := 1 to 8 do
        begin
          With ElevStats.Stats[ind] do
            begin
              ElevStats.N[CC] := 0;
              Min[CC] := 1e6;
              Max[CC] := 1e-6;
              Sum[CC] := 0;
              SumX2[CC] := 0;
              Sum_e2[CC] := 0;
              Mean[CC] := 0;
              StDev[CC] := 0;
            end;
          for j := 0 to HistWidth-1 do
            ElevStats.Histogram[ind,CC,j] := 0;
        end;
        ElevStats.pLowerMin[CC] := 0;
    end;

  // Show Progress Form
  ProgForm.Show;

  // Progress bar message
  ProgForm.ProgressLabel.Caption := 'Calculating Elevation Stats.';

  // Collecting data for each row and column
  FOR ER := 0 To (Site.Rows-1) Do
    FOR EC := 0 To (Site.Cols-1) Do
      Begin  {First Pass Statistics}

        // Updating the progress bar
        If EC=0 then ProgForm.Update2Gages(Trunc(ER/Site.Rows*(100)));  //FIXME USERSTOP

        // Get the information of the cell
        RetA(ER, EC, StatCell);

        // If cell is diked do not include it in the elevation statistics
        If ClassicDike and StatCell.ProtDikes then Continue;
        if (not ClassicDike) and (StatCell.ProtDikes) and (not StatCell.ElevDikes) then Continue;


        // If the subsite has preprocessed data do not include the cell in the elevation statistics
        SubSite := Site.GetSubSite(EC,ER,@StatCell);
        If SubSite.Use_PreProcessor then Continue;

        // If the GT=0 do not include in the elevation statisitcs
        if subsite.MHHW=0 then Continue;

        // Get cell category
        CC := GetCellCat(@StatCell);

        // Adjustment of the slope
        SlopeAdjustment := (Site.Scale*0.5) * StatCell.TanSlope;   {QA 11.10.2005}

        // Get elevation category
        CE := CatElev(@StatCell,CC);

        // If elevation category is 999 (no data) do not process
        if CE = 999 then continue;

        // Then elevation in m is ...
        if CC = Blank then ElevLevel[5] := 0
                     else ElevLevel[5] := (CE+SlopeAdjustment);

        ElevLevel[6] := ElevLevel[5] - subsite.MHHW;
        ElevLevel[7] := ElevLevel[5] - subsite.MLLW;
        ElevLevel[8] := ElevLevel[5] + StatCell.MTLminusNAVD;

        // Elevation in HTU
        ElevLevel[1] := ElevLevel[5] / subsite.MHHW;
        ElevLevel[2] := ElevLevel[6] / subsite.MHHW;
        ElevLevel[3] := ElevLevel[7] / subsite.MHHW;
        ElevLevel[4] := ElevLevel[8] / subsite.MHHW;

        // Get subsite wetland max and min elevations
        MinWetElev := LowerBound(CC,SubSite);
        if ElevLevel[5]<MinWetElev then ElevStats.pLowerMin[CC] := ElevStats.pLowerMin[CC]+1;

        // Update the statistics
        With ElevStats do
          Begin
            // Increment the category by one
            Inc(N[CC]);

            // Update minimum, maximum, sum, squared sum
            for ind := 1 to 8 do
              With Stats[ind] do
                Begin
                  If ElevLevel[ind] < Min[CC] then Min[CC] := ElevLevel[ind];
                  If ElevLevel[ind] > Max[CC] then Max[CC] := ElevLevel[ind];
                  Sum[CC] := Sum[CC] + ElevLevel[ind];
                  SumX2[CC] := SumX2[CC] + ElevLevel[ind]*ElevLevel[ind];
                End;
          End;
      End;

  for CC := Firstcategory to Lastcategory do
    begin
      With ElevStats do
        begin
          For ind := 1 to 8 do
            begin
              //Inizialize arrays
              SetLength(Values[ind,CC],N[CC]);

              // Calculating the mean for each category
              With Stats[ind] do
                If N[CC] > 0 then Mean[CC] := Sum[CC]/N[CC]
                        else Mean[CC] := 0;
            end;

          //Fraction of cells below min elevation
          If N[CC] > 0 then pLowerMin[CC] := 100*pLowerMin[CC]/N[CC]
                        else pLowerMin[CC] := 0;

          // Reinizialize the number of cells in a particualr wetland type
          N[CC] := 0;
        end;
    end;

  // Second pass once the dimensions for Values and Val_M are obtained
  for ER := 0 to (Site.Rows-1) Do
    for EC := 0 to (Site.Cols-1) Do
      begin  {Second Pass Statistics}

        // Progress bar updates
        If EC=0 then ProgForm.Update2Gages(Trunc(ER/Site.Rows*(100)));  //FIXME USERSTOP

        // Get the information of the cell
        RetA(ER, EC, StatCell);

        // If cell is diked do not include it in the elevation statistics
        If ClassicDike and StatCell.ProtDikes then Continue;
        if (not ClassicDike) and (StatCell.ProtDikes) and (not StatCell.ElevDikes) then Continue;

        // If the subsite has preprocessed data do not include the cell in the elevation statistics
        SubSite := Site.GetSubSite(EC,ER,@StatCell);
        If SubSite.Use_PreProcessor then Continue;  {Don't calculate elevations from Pre-Processed Data}

        // If the GT=0 do not include in the elevation statisitcs
        if subsite.MHHW=0 then Continue;

        // Get cell category
        CC := GetCellCat(@StatCell);

        // Calculate slope
        SlopeAdjustment := (Site.Scale*0.5) * StatCell.TanSlope;   {QA 11.10.2005}

        // Get cell elevation
        CE:= CatElev(@StatCell,CC);

        // If elevation category is 999 (no data) do not process
        if CE = 999 then continue;  // dont process blank elevs

        // Elevation in m is ...
        //If CC= Blank then ElevLevel[2] := 0
        //            else ElevLevel[2] := (CE+SlopeAdjustment);

        // Elevation in HTU
        //ElevLevel[1] := ElevLevel[2] / subsite.MHHW;

        // Then elevation in m is ...
       // Then elevation in m is ...
       if CC = Blank then ElevLevel[5] := 0
                     else ElevLevel[5] := (CE+SlopeAdjustment);

       ElevLevel[6] := ElevLevel[5] - subsite.MHHW;
       ElevLevel[7] := ElevLevel[5] - subsite.MLLW;
       ElevLevel[8] := ElevLevel[5] + StatCell.MTLminusNAVD;

       // Elevation in HTU
       ElevLevel[1] := ElevLevel[5] / subsite.MHHW;
       ElevLevel[2] := ElevLevel[6] / subsite.MHHW;
       ElevLevel[3] := ElevLevel[7] / subsite.MHHW;
       ElevLevel[4] := ElevLevel[8] / subsite.MHHW;

        // Update the statistics
        With ElevStats do
          Begin
            // Increment the category by one
            Inc(N[CC]);

            for ind := 1 to 8 do
              begin
                // Storing elevation values
                Values[ind,CC,N[CC]-1] := ElevLevel[ind];

                With Stats[ind] do
                  begin
                    // Updating standard deviations
                    Sum_e2[CC] := Sum_e2[CC] + Sqr(ElevLevel[ind]-Mean[CC]);

                    //Update the elevation histogram
                    If (ElevLevel[ind] >= Min[CC]) and (ElevLevel[ind]<Max[CC]) then
                      Inc(Histogram[ind,CC,TRUNC(HistWidth*(ElevLevel[ind]-Min[CC])/(Max[CC]-Min[CC]))]);  {increment appropriate bin}
                  end;
              end;
          end;
      end;

  for CC := Firstcategory to Lastcategory do
    with ElevStats do
      begin
        for ind := 1 to 8 do
          with Stats[ind] do
            begin
              if N[CC]<=1
                then begin StDev[CC] := 0; P05[CC]:=0; P95[CC]:=0; end
              else
                begin
                  // Calculate standard deviation
                  StDev[CC] := sqrt(Sum_e2[CC]/(n[CC]-1));
                  SortVals := Values;

                  // Sorting values
                  QuickSort(SortVals[ind,CC],0,N[CC]-1);

                  // Calculate 5th percentile
                  PctInd :=Round((N[CC]+1)*0.05)-1;
                  If PctInd<0 then PctInd := 0;
                  P05[CC] := SortVals[ind,CC,PctInd];

                  // Calculate 95th percentile
                  PctInd :=Round((N[CC]+1)*0.95)-1;
                  If (PctInd>N[CC]-1) then PctInd := N[CC]-1;
                  P95[CC] := SortVals[ind,CC,PctInd];
                end;
            end;
      end;

   // Deallocate dynamic arrays
   for CC := Firstcategory to Lastcategory do
    for ind := 1 to 8 do
      begin
        Values[ind,CC] := nil;
      end;

   Init_ElevStats := True;
   ElevStats.Derived := Now();
   ProgForm.Hide;

End;


{-----------------------------------------------------------------------------------}

Procedure TSLAMM_Simulation.Calc_Elev_Stats_2;
//============================================
// Calculation of elevation statistics
//============================================
Var
    SlopeAdjustment: Double; {elev stat code}
//    ElevFile: TextFile;
    PctInd, Ind,ER,EC, j: Integer;
    CC: Category;
    StatCell: CompressedCell;
    SubSite: TSubSite;
//    Values, SortVals: Array[1..8,Category] of Array of Single;
    CE: Double;
    ElevLevel: array[1..8] of Double;
    MinWetElev: Double;
    NCount, i : integer;

Begin      {Map must be loaded for this calculation}

  // Initialize statistical indicators
  for CC := FirstCategory to Lastcategory do
    begin
      for Ind := 1 to 8 do
        begin
          With ElevStats.Stats[ind] do
            begin
              ElevStats.N[CC] := 0;
              Min[CC] := 1e6;
              Max[CC] := 1e-6;
              Sum[CC] := 0;
              SumX2[CC] := 0;
              Sum_e2[CC] := 0;
              Mean[CC] := 0;
              StDev[CC] := 0;
            end;
          for j := 0 to HistWidth-1 do
            ElevStats.Histogram[ind,CC,j] := 0;
        end;
        ElevStats.pLowerMin[CC] := 0;
    end;

  // Show Progress Form
  ProgForm.Show;

  // Progress bar message
  ProgForm.ProgressLabel.Caption := 'Calculating Elevation Stats.';

  // Collecting data for each row and column
  FOR ER := 0 To (Site.Rows-1) Do
    FOR EC := 0 To (Site.Cols-1) Do
      Begin  {First Pass Statistics}

        // Updating the progress bar
        If EC=0 then ProgForm.Update2Gages(Trunc(ER/Site.Rows*(100)));  //FIXME USERSTOP

        // Get the information of the cell
        RetA(ER, EC, StatCell);

        // If cell is diked do not include it in the elevation statistics
        If ClassicDike and StatCell.ProtDikes then Continue;
        if (not ClassicDike) and (StatCell.ProtDikes) and (not StatCell.ElevDikes) then Continue;

        // If the subsite has preprocessed data do not include the cell in the elevation statistics
        SubSite := Site.GetSubSite(EC,ER,@StatCell);
        If SubSite.Use_PreProcessor then Continue;

        // If the GT=0 do not include in the elevation statisitcs
        if subsite.MHHW=0 then Continue;

        // Get cell category
        CC := GetCellCat(@StatCell);

        // Adjustment of the slope
        SlopeAdjustment := (Site.Scale*0.5) * StatCell.TanSlope;   {QA 11.10.2005}

        // Get elevation category
        CE := CatElev(@StatCell,CC);

        // If elevation category is 999 (no data) do not process
        if CE = 999 then continue;

        // Then elevation in m is ...
        if CC = Blank then ElevLevel[5] := 0
                     else ElevLevel[5] := (CE+SlopeAdjustment);

        ElevLevel[6] := ElevLevel[5] - subsite.MHHW;
        ElevLevel[7] := ElevLevel[5] - subsite.MLLW;
        ElevLevel[8] := ElevLevel[5] + StatCell.MTLminusNAVD;

        // Elevation in HTU
        ElevLevel[1] := ElevLevel[5] / subsite.MHHW;
        ElevLevel[2] := ElevLevel[6] / subsite.MHHW;
        ElevLevel[3] := ElevLevel[7] / subsite.MHHW;
        ElevLevel[4] := ElevLevel[8] / subsite.MHHW;

        // Get subsite wetland max and min elevations
        MinWetElev := LowerBound(CC,SubSite);
        if ElevLevel[5]<MinWetElev then ElevStats.pLowerMin[CC] := ElevStats.pLowerMin[CC]+1;

        // Update the statistics
        With ElevStats do
          Begin
            // Increment the category by one
            Inc(N[CC]);

            // Update minimum, maximum, sum, squared sum
            for ind := 1 to 8 do
              With Stats[ind] do
                Begin
                  If ElevLevel[ind] < Min[CC] then Min[CC] := ElevLevel[ind];
                  If ElevLevel[ind] > Max[CC] then Max[CC] := ElevLevel[ind];
                  Sum[CC] := Sum[CC] + ElevLevel[ind];
                  SumX2[CC] := SumX2[CC] + ElevLevel[ind]*ElevLevel[ind];
                End;
          End;
      End;

  for CC := Firstcategory to Lastcategory do
    begin
      With ElevStats do
        begin
          For ind := 1 to 8 do
            begin
              //Inizialize arrays
//              SetLength(Values[ind,CC],N[CC]);

              // Calculating the mean for each category
              With Stats[ind] do
                If N[CC] > 0 then Mean[CC] := Sum[CC]/N[CC]
                        else Mean[CC] := 0;
            end;

          //Fraction of cells below min elevation
          If N[CC] > 0 then pLowerMin[CC] := 100*pLowerMin[CC]/N[CC]
                        else pLowerMin[CC] := 0;

          // Reinizialize the number of cells in a particualr wetland type
          N[CC] := 0;
        end;
    end;

  // Second pass once the dimensions for Values and Val_M are obtained
  for ER := 0 to (Site.Rows-1) Do
    for EC := 0 to (Site.Cols-1) Do
      begin  {Second Pass Statistics}

        // Progress bar updates
        If EC=0 then ProgForm.Update2Gages(Trunc(ER/Site.Rows*(100)));  //FIXME USERSTOP

        // Get the information of the cell
        RetA(ER, EC, StatCell);

        // If cell is diked do not include it in the elevation statistics
        If ClassicDike and StatCell.ProtDikes then Continue;
        if (not ClassicDike) and (StatCell.ProtDikes) and (not StatCell.ElevDikes) then Continue;

        // If the subsite has preprocessed data do not include the cell in the elevation statistics
        SubSite := Site.GetSubSite(EC,ER,@StatCell);
        If SubSite.Use_PreProcessor then Continue;  {Don't calculate elevations from Pre-Processed Data}

        // If the GT=0 do not include in the elevation statisitcs
        if subsite.MHHW=0 then Continue;

        // Get cell category
        CC := GetCellCat(@StatCell);

        // Calculate slope
        SlopeAdjustment := (Site.Scale*0.5) * StatCell.TanSlope;   {QA 11.10.2005}

        // Get cell elevation
        CE:= CatElev(@StatCell,CC);

        // If elevation category is 999 (no data) do not process
        if CE = 999 then continue;  // dont process blank elevs

        // Elevation in m is ...
        //If CC= Blank then ElevLevel[2] := 0
        //            else ElevLevel[2] := (CE+SlopeAdjustment);

        // Elevation in HTU
        //ElevLevel[1] := ElevLevel[2] / subsite.MHHW;

        // Then elevation in m is ...
       // Then elevation in m is ...
       if CC = Blank then ElevLevel[5] := 0
                     else ElevLevel[5] := (CE+SlopeAdjustment);

       ElevLevel[6] := ElevLevel[5] - subsite.MHHW;
       ElevLevel[7] := ElevLevel[5] - subsite.MLLW;
       ElevLevel[8] := ElevLevel[5] + StatCell.MTLminusNAVD;

       // Elevation in HTU
       ElevLevel[1] := ElevLevel[5] / subsite.MHHW;
       ElevLevel[2] := ElevLevel[6] / subsite.MHHW;
       ElevLevel[3] := ElevLevel[7] / subsite.MHHW;
       ElevLevel[4] := ElevLevel[8] / subsite.MHHW;

        // Update the statistics
        With ElevStats do
          Begin
            // Increment the category by one
            Inc(N[CC]);

            for ind := 1 to 8 do
              begin
                // Storing elevation values
                //Values[ind,CC,N[CC]-1] := ElevLevel[ind];

                With Stats[ind] do
                  begin
                    // Updating standard deviations
                    Sum_e2[CC] := Sum_e2[CC] + Sqr(ElevLevel[ind]-Mean[CC]);

                    //Update the elevation histogram
                    If (ElevLevel[ind] >= Min[CC]) and (ElevLevel[ind]<Max[CC]) then
                      Inc(Histogram[ind,CC,TRUNC(HistWidth*(ElevLevel[ind]-Min[CC])/(Max[CC]-Min[CC]))]);  {increment appropriate bin}
                  end;
              end;
          end;
      end;

  for CC := Firstcategory to Lastcategory do
    with ElevStats do
      begin
        for ind := 1 to 8 do
          with Stats[ind] do
            begin
              if N[CC]<=1
                then begin StDev[CC] := 0; P05[CC]:=0; P95[CC]:=0; end
              else
                begin
                  // Calculate standard deviation
                  StDev[CC] := sqrt(Sum_e2[CC]/(n[CC]-1));

                  //SortVals := Values;

                  // Sorting values
                  //QuickSort(SortVals[ind,CC],0,N[CC]-1);

                  // Calculate 5th percentile
                  PctInd :=Round((N[CC]+1)*0.05)-1;
                  If PctInd<0 then PctInd := 0;
                  P05[CC] := 0;
                  NCount :=0;
                  i := 0;
                  while NCount < PctInd do
                    begin
                      NCount := NCount + Histogram[ind,CC,i];
                      i := i+1;
                    end;
                  P05[CC] := (((Max[CC]-Min[CC])/HistWidth)*(i-1))+ Min[CC] + ((Max[CC]-Min[CC])/HistWidth)*0.5;

                  //SortVals[ind,CC,PctInd];

                  // Calculate 95th percentile
                  PctInd :=Round((N[CC]+1)*0.95)-1;
                  If (PctInd>N[CC]-1) then PctInd := N[CC]-1;
                  P95[CC] := 0;
                  NCount :=0;
                  i := 0;
                  while NCount < PctInd do
                    begin
                      NCount := NCount + Histogram[ind,CC,i];
                      i := i+1;
                    end;
                  P95[CC] := (((Max[CC]-Min[CC])/HistWidth)*(i-1)) + Min[CC] + ((Max[CC]-Min[CC])/HistWidth)*0.5;

                end;
            end;
      end;

   // Deallocate dynamic arrays
{   for CC := Firstcategory to Lastcategory do
    for ind := 1 to 8 do
      begin
        Values[ind,CC] := nil;
      end;
}
   Init_ElevStats := True;
   ElevStats.Derived := Now();
   ProgForm.Hide;

End;



FUNCTION TSLAMM_Simulation.RunOneYear(IsT0: Boolean): Boolean;

    {------------------------------------------------------------}
    {Calc Fractal Dimension and Shoreline Protection Potential}
       Procedure Calc_FractalNos;
       Var i, EC,ER: Integer;
           ProcCell: CompressedCell;
       Begin
         For i := 0 to Site.NOutputSites  do  {No ROS for now}
           FRs[i].ShoreProtect := 0;

         ProgForm.ProgressLabel.Caption:='Calculating Fractal Dimension';
         ProgForm.Update;

         Result := ChngWater(False,False);
         If Result then
          For ER := 0 to (Site.Rows-1) Do
           FOR EC := 0 To (Site.Cols-1) Do
            Begin
              If EC=1 then
                Begin
                  Result := ProgForm.Update2Gages(Trunc((Er)/Site.Rows*100));
                  If Not Result then Exit;
                End;
              RetA(ER,EC,ProcCell);
              FractalNos(ER,EC,ProcCell,Site);  {collect info for fractal dimension of init.cond}
            End; {For Do Loops}
       End;
    {------------------------------------------------------------}

Var ER,EC: Integer;
    AddStr: String;
    ProcCell: CompressedCell;
    ERODELOOP: Boolean;
    ProgStr : String;

    ReadCell: CompressedCell;
    CC: category;
    NWI_Number: integer;
    // i : integer;
    // DevDrySet: Boolean;

Begin  {RunOneYear}

  If Not IsT0
    then
      Begin
        ProgStr := 'Running Yearly Scenario : ';
        ProgForm.YearLabel.Visible:=True;
      End
    else
      Begin
        ProgStr := 'Running Time Zero : ';
        ProgForm.YearLabel.Visible:=False;
      End;


  ProgForm.ProgressLabel.Caption:=ProgStr;
  ProgForm.Show;
  LocalSLChange(Site, Year, IsT0);  // Updates the "Year" Variable and Local SL Variables

  // Check for annual maps if it is not time zero
  if (not IsT0) then CheckForAnnualMaps;

(*
  // Quick and dirty conversion of the cells for TNC Walsh - #1-2 Green Scenarios - MARCO
  if Year = 2025 then
    begin
      for ER:=0 to Site.Rows-1 do
        begin
          for EC:=0 to Site.Cols-1 do
            begin
              if (ROSArray[ER*Site.Cols+EC] = 1) or (ROSArray[ER*Site.Cols+EC] = 3)   then
                begin
                  // Get the cell object
                  RetA(ER,EC,ReadCell);

                  // DVLP2UNDVLP
                  for i := 1 to NUM_CAT_COMPRESS do
                    begin
                      if ReadCell.Cats[i] = DevDryLand then
                        begin
                          ReadCell.Cats[i] := UndDryLand;
                          break;
                        end;
                    end;

                  // Set cell attributes
                  SetA(ER,EC,ReadCell);
                end;
            end;
        end;
    end;
*)




  ProgForm.YearLabel.Caption:=IntToStr(Year);
  Result := UpdateElevations;
  If Not Result then Exit;

  Result := ChngWater(True,Not IsT0);
  If not Result then Exit;

  if CheckConnectivity then Result := Calc_Connectivity(-1,-1);
  If not Result then Exit;

  if (not isT0) then
    if (trim(SalFileN)='') or (Pos('.xls',lowercase(SalFileN))>0) then
          Result := CalcSalinity(false,false)  //not using raster salinity file so try to calc salinity if relevant
     else Result := True;

  if not Result then Exit;

  GridForm.Image1.Canvas.Pen.Mode := pmcopy;

 { for processors := 1 to maxprocessors do
    transferthread := TTransferthread.create(minrow,maxrow)}

  FOR ERODELOOP := FALSE to TRUE Do
   FOR ER := 0 To (Site.Rows-1) Do
    FOR EC := 0 To (Site.Cols-1) Do
     Begin
       If (EC mod 400) =0 then
         begin
           If ErodeLoop then AddStr := 'Erosion'
                        else AddStr := 'Inundation';
           ProgForm.ProgressLabel.Caption:=ProgStr+AddStr;
           Result := ProgForm.Update2Gages(Trunc((Er)/Site.Rows*100));
           If Not Result then Exit;
         end;

       EachRow:=ER;  EachCol:=EC;
       RetA(ER, EC, ProcCell);
       Transfer(@ProcCell, ER, EC, ERODELOOP);
       SetA(ER, EC, ProcCell);

       If ERODELOOP then
        If Display_Screen_Maps then
         If Gridform.IsVisible then
           Gridform.DrawCell(@ProcCell,ER,EC,GetCellCat(@ProcCell),False,False,True,0);
     End; {For Do Loops}

   Calc_FractalNos;
   If Not Result then Exit;

   Inc(CountIter);
   Summarize(Map,Tropical,Summary,Year);  {write array for tabular output for first year}

   // PROGFORM.Hide;
 End; {Run One Year}


{------------------------------------------------------------------------------}

Procedure TSLAMM_Simulation.ExecuteRun;
Var Prot_Scenario: ProtScenario;
    IPCC_Scenario: IPCCScenarios;
    IPCC_Est     : IPCCEstimates;

          {==============================================================================}
          Procedure InitSubSite(PSS: TSubSite);
          Begin
           With PSS do
            Begin
              Norm := 0;
              NewSL := 0.0;
              OldSL := 0.0;
              SLRise := 0.0;
              DeltaT := 0;
              LastOverwash := Year;
            End;
          End;
          {==============================================================================}

    {----------------------------------------------------------------------------------------}
    Function InitYearSLAM: Boolean;
    Var
      i,j: integer;
      CC: Category;
      ER,EC: Integer;

          function CharOccurs(const str: string; c: char): integer;
          // Returns the number of times a character occurs in a string
          // Ernesto De Spirito, http://www.delphi3000.com/articles/article_2615.asp
          var
            p: PChar;
          begin
            Result := 0;
            p := PChar(Pointer(str));
            while p <> nil do begin
              p := StrScan(p, c);
              if p <> nil then begin
                inc(Result);
                inc(p);
              end;
            end;
          end;
          {==============================================================================}

    Var NumRows: Integer;
        MR: TModalResult;
        ProcCell: CompressedCell;

    BEGIN  {InitYearSLAM}
      Result := True;

      Year := 0;  {Output as Init Cond}
      IF SaveGIS Then If GISForm.OutputYear(Year) THEN Result := Save_GIS_Files;
      If Not Result then Exit;  //user cancel button

      Year := Site.T0;  {latest NWI Photo Date for all Subsites}

      CatSums := nil;
      SetLength(CatSums, Site.NOutputSites+1+Site.MaxROS );

      Summary := nil;
      IF RunSpecificYears
        then NumRows := 3 + CharOccurs(YearsString,',')    //numrows = 3 + number of commas in comma delimited string
        else NumRows := (4+( (MaxYear-2000) Div TimeStep));

      SetLength(FRs,Site.NOutputSites+1); {Fractal Numbers Data}
      SetLength(Summary,Site.NOutputSites+1+Site.MaxROS);

      For i := 0 to Site.NOutputSites+Site.MaxROS  do
        Begin
          FillChar(CatSums[i], SizeOf(CatSums[i]),'0');
          SetLength(Summary[i],NumRows);
          For j := 0 to NumRows-1 do
            FillChar(Summary[i,j], SizeOf(Summary[i,j]),'0');
        End;

      SetLength(RowLabel,(Site.NOutputSites+1+Site.MaxROS)*NumRows);
      For i := 1 to Length(RowLabel) do
        RowLabel[i-1] := '';

      CellHa := Site.Scale * Site.Scale * 0.0001;
                                        {/ 10000 }

      ProtectDeveloped :=(Prot_Scenario>NoProtect);     //use global loop variable
      ProtectAll :=(Prot_Scenario>ProtDeveloped);

      FOR CC := FirstCategory TO LastCategory DO
        FracLost[CC] := -9999;   {ensure correct initialization}

      InitSubSite(Site.GlobalSite);
      For i := 0 to Site.NSubSites-1 do
        InitSubSite(Site.SubSites[i]);

      CountIter := 0;

      If (Length(BMatrix) < Site.Rows * Site.Cols) then
        Begin
          BMatrix := nil;
          SetLength (BMatrix,Site.Rows * Site.Cols);
        End;
      For i:=0 to (Site.Rows-1) do
        For j:=0 to (Site.Cols-1) do
          BMatrix[Site.Cols*(i)+(j)]:=0;

      For i := 0 to Site.NOutputSites+Site.MaxROS  do
        FillChar(CatSums[i], SizeOf(CatSums[i]),0);

      ColLabel[0] := 'SLR (eustatic)';
      FOR CC := FirstCategory to LastCategory DO
        ColLabel[ORD(CC)+1] := TitleCat[CC];
      ColLabel[ORD(LastCategory)+2] := 'FractalD';
      ColLabel[ORD(LastCategory)+3] := 'FD. SM. Only';
      ColLabel[ORD(LastCategory)+4] := 'Shore Prot. (km)';
      ColLabel[ORD(LastCategory)+5] := 'Aggregated Non Tidal';
      ColLabel[ORD(LastCategory)+6] := 'Freshwater Non-Tidal';
      ColLabel[ORD(LastCategory)+7] := 'Open Water';
      ColLabel[ORD(LastCategory)+8] := 'Low Tidal';
      ColLabel[ORD(LastCategory)+9] := 'Saltmarsh';
      ColLabel[ORD(LastCategory)+10] := 'Transitional';
      ColLabel[ORD(LastCategory)+11] := 'Freshwater Tidal';

      For j := 0 to Site.NOutputSites  do
       With FRs[j] do
        Begin
          For i := 0 to 1 do
            Begin
              MaxFDRow[i] := 0;  {reinitialize fractal dimension variables}
              MaxFDCol[i] := 0;
              MinFDRow[i] := 65535;
              MinFDCol[i] := 65535;
              NBox[i]     := 0;
            End;
          ShoreProtect := 0;
        End;

{      InitElevRanges; }

      Result := ProgForm.Update2Gages(0);
      If Not Result then Exit;
      
      ProgForm.YearLabel.Visible:=True;
      ProgForm.YearLabel.Caption:=IntToStr(Year);
      ProgForm.Show;

      GridForm.Image1.Canvas.Pen.Mode := pmcopy;
      Result :=  ChngWater(False,False); {update water attributes for the entire map, just for initial conditions summary}
      If Not Result then Exit;

      FOR ER := 0 To (Site.Rows-1) Do
       FOR EC := 0 To (Site.Cols-1) Do
        Begin  {Set CatSums Initial Condition}
          RetA(ER, EC, ProcCell);
          For i := 0 to Site.NOutputSites do
            For CC := FirstCategory to LastCategory Do
              If (i=0) or Site.InOutSite(EC,ER,i) then
                CatSums[i,CC] := CatSums[i,CC] + (CellWidth(@ProcCell,CC) * Site.Scale);
                  {m2}                {m2}             {m}                    {m}
          For i := 1 to Site.MaxROS do
            For CC := FirstCategory to LastCategory Do
              If (i=ROSArray[(Site.Cols*ER)+EC]) then
                CatSums[i+Site.NOutputSites,CC] := CatSums[i+Site.NOutputSites,CC] + (CellWidth(@ProcCell,CC) * Site.Scale);
                                           {m2}                                {m2}                       {m}         {m}


          FractalNos(ER,EC,ProcCell,Site);  {collect info for fractal dimension of init.cond}

          If (EC mod 400) =0 then
            begin
              ProgForm.ProgressLabel.Caption:='Collecting Init. Cond Data';
              ProgForm.ProgressLabel.Visible := True;
              Result := ProgForm.Update2Gages(Trunc((Er)/Site.Rows*100));
              If Not Result then Exit;
            end;
        End;

      ProgForm.ProgressLabel.Caption:='Running First Year';

//    CALC_ELEV_STATS;

      Summarize(Map,Tropical,Summary,0);
               {Show initial conditions and define Tropical and AreaWater}

      For i := 0 to Site.NOutputSites + Site.MaxROS  do
        FillChar(CatSums[i], SizeOf(CatSums[i]),0);  {zero out catsums}

      If RunFirstYear Then Result := RunOneYear(True);
      If Not Result then Exit;

      IF SaveGIS Then If GISForm.OutputYear(Year) THEN Result := Save_GIS_Files;
      If Not Result then Exit;

      If GridForm.Visible then
       With GridForm do
        Begin
          SS := Self;
          IF RunFirstYear then
            Begin
              If QA_Tools then  {display map after "time zero" run}
                 Begin
                   hide; MapTitle.Visible := True;
                   MapTitle.Caption:= 'Map at Year '+IntToStr(Year);
                   UserStop := False;
                   RunPanel.Visible := False;
                   MR := ShowModal;

                   IF (MR=2) then if MessageDlg('Stop Simulation Execution And Return to Main Menu? ',mtconfirmation,[mbyes,mbno],0) = mryes then
                     MR := MRAbort;   // Disambiguation of X key

                   If MR = MRAbort then Begin Result := False; UserStop := True; exit; End;
                   If MR = MRIgnore then QA_Tools := False;
                   Show; Update; MapTitle.Visible := False; IsVisible := True;
                 end;

              If Maps_to_MSWord or Maps_to_GIF then Result := CopyMapsToMSWord;

              If Not Result then Exit;
              ShowRunPanel;

             End; {runfirstyear}

          Result := RunFirstYear;
        End;

    END; {InitYearSLAM}
    {------------------------------------------------------------}
    {------------------------------------------------------------}
    {------------------------------------------------------------}
    Function CheckValidSLAMM: Boolean;
    Var j: Integer;
        NoPoly: Boolean;
    Begin
      Result := ValidateOutputSites;
      If Not Result then Exit;

        Begin   // Study Validation Routine
          For j := 0 to NumFWFlows-1 do
            Begin
              NoPoly := False;
              If FWFlows[j].Poly = nil then NoPoly := True
                 else If FWFlows[j].Poly.NumPts < 3 then NoPoly := True;

              If NoPoly then
                Begin
                  MessageDlg('Error, Fresh water flow "'+FWFlows[j].Name +'" extent is not defined.',mterror,[mbok],0);
                  Result := False;
                  Exit;
                End;

              If (FWFlows[j].ExtentOnly = False) and (FWFlows[j].NumSegments = 0) then
                Begin
                  MessageDlg('Error, Fresh water flow "'+FWFlows[j].Name +'" origin, mouth is not defined.',mterror,[mbok],0);
                  Result := False;
                  Exit;
                End;
            End;
        End;
    End;
    {-------------------------------------------------------------------------------------------}

  Function ExecuteSlamm(ProtScen:ProtScenario;
                          IPCCScen: IPCCScenarios;IPCCEst: IPCCEstimates; FixNum: Integer): Boolean;
                           {Execute a single iteration of SLAMM}
   Var  i : Integer;
        MR: TModalResult;
   BEGIN

     Result := CheckValidSLAMM;
     If Not Result then Begin UserStop := True; Exit; End;

     Gridform.GraphBox.ItemIndex := 0;
     For i := 0 to NumFWFlows-1 do
       FWFLows[i].RetentionInitialized := False;


    If SaveMapsToDisk and Not MakeNewDiskMap
      then if not FileExists(S60FileN) then
         Raise ESlammError.Create('Data File "'+S60FileN+'" Does Not Exist!');

    Result := MakeDataFile(False,'','');
     If Not Result then Begin UserStop := True; Exit; End;

    Site.InitElevVars;

    If SaveMapsToDisk then
      Begin
        Try
          DisposeFile(MapHandle,False,BackupFN);
        Except
          {oops, it's already closed}
        End;

        If Not CopyFileProg(S60FileN, BackupFN) then
          Begin UserStop := True; Exit; End;  {Copy file to *.BAK for working with}
        LoadFile(MapHandle,BackupFN);   {Work with Backup File}
      End;

    If Display_Screen_Maps then
       With GridForm do
        Begin   {Work Around what appears to be a screen draw bug in Delphi 4.0}
          PageControl1.ActivePageIndex := 1;
          PageControl1.Pages[0].TabVisible := False;
          PageControl1.Pages[2].TabVisible := False;
          PageControl1.Pages[3].TabVisible := False;

          PageControl1.Pages[0].Caption := 'Analysis Tools';

          HaltButton.Visible := True;
          NextStepButton.Visible := True;
          StopPausingButton.Visible := True;

          Image1.Height:=Site.Rows+2;
          Image1.Width:=Site.Cols+2;
          Image1.Picture.Bitmap.PixelFormat := pf24bit;  // Do this first

          ZoomFactor := InitZoomFactor;

          IsVisible:=True;

          Running_Fixed := (FixNum>0) and (FixNum<4);
          Running_Custom := (FixNum>3);

          FixedScen := FixNum;
          IPCCSLRate:= IPCCScen;
          IPCCSLEst:= IPCCEst;
          ProtectDeveloped :=(Prot_Scenario>NoProtect);
          ProtectAll :=(Prot_Scenario>ProtDeveloped);

          GridForm.Caption := 'SLAMM 6 Map -- '+ExtractFileName(FileN);
          RunPanel.Visible := False;

          Result := ShowMap(Site.Rows,Site.Cols,Site,Self);
          If (Not Result) or (UserStop) then Begin Result := False; UserStop:=True; Exit; End;

          Year := 0;
          ShowRunPanel;
        End;

    Result := PreProcessWetlands;
    If Not Result then Begin UserStop := True; Exit; End;

    If IncludeDikes then InitializeDikes;

    Year :=  Site.GlobalSite.NWI_Photo_Date;
    InitSubSite(Site.GlobalSite);
    For i := 0 to Site.NSubSites-1 do
      InitSubSite(Site.SubSites[i]);

{    if (trim(SalFileN)<>'') or (Pos('.xls',lowercase(SalFileN))=0)
         then Result := True  // don't calculate salinity, a file is being used instead.
         else } Result := CalcSalinity(true,True);

    If Not Result then Begin UserStop := True; Exit; End;

    If GridForm.Visible then  {display map after pre-processor and salinity calculation}
     With GridForm do
      Begin
        SS := Self;
        If QA_Tools then
           Begin ProgForm.YearLabel.Visible:=True;
                 ProgForm.YearLabel.Caption:= 'Init.Cond';
                 ProgForm.Show;

                 hide; MapTitle.Visible := True;
                 MapTitle.Caption:= 'NWI Processed Elevations Over NED';
                 UserStop := False;
                 RunPanel.Visible := False;
                 MR := ShowModal;

                 IF (MR=2) then if MessageDlg('Stop Simulation Execution And Return to Main Menu? ',mtconfirmation,[mbyes,mbno],0) = mryes then
                     MR := MRAbort;   // Disambiguation of X key

                 If MR = MRAbort then Begin Result := False; UserStop:=True; exit; End;
                 If MR = MRIgnore then QA_Tools := False;
                 Show; Update; MapTitle.Visible := False; IsVisible := True;
           end;

        Year := 0;  // label maps as initial condition
        If Maps_to_MSWord or Maps_to_GIF then Result := CopyMapsToMSWord; // 12/15/09 Wait for pre-processing & Salinity Calcs

        If Not Result then Begin UserStop := True; Exit; End;

        Year :=  Site.GlobalSite.NWI_Photo_Date;  // return to current date

        ShowRunPanel;

      End;

    ProgForm.YearLabel.Visible       :=False;
    ProgForm.SLRLabel.Visible        := True;
    ProgForm.ProtectionLabel.Visible := True;
    If (FixNum>0) and (FixNum<4)
                then ProgForm.SLRLabel.Caption  := LabelFixed[FixNum]
    else if FixNum > 3 then ProgForm.SLRLabel.Caption  := FloatToStrF(CustomSLR ,ffgeneral,4,2)+ ' meters '
                else ProgForm.SLRLabel.Caption  := LabelIPCC[IPCCScen]+' '+LabelIPCCEst[IPCCEst];

    ProgForm.ProtectionLabel.Caption := LabelProtect[ProtScen];

    ProgForm.Show;

    Running_Fixed := (FixNum>0) and (FixNum<4);
    Running_Custom := (FixNum>3);
    FixedScen := FixNum;
    IPCCSLRate:= IPCCScen;
    IPCCSLEst:= IPCCEst;

    Result := InitYearSLAM;  {RUN FIRST YEAR HERE}
    If Not Result then Begin UserStop:=True; Exit; End;

    REPEAT  {Run additional Year}
      For i := 0 to Site.NOutputSites+Site.MaxROS  do
         FillChar(CatSums[i], SizeOf(CatSums[i]),0);

      Result := RunOneYear(False);
      If Not Result then Begin UserStop:=True; Exit; End;

      IF SaveGIS Then If GISForm.OutputYear(Year) THEN Result := Save_GIS_Files;
      If Not Result then Begin UserStop:=True; Exit; End;

    If GridForm.Visible then
    With GridForm do
     Begin
       SS := Self;
       If QA_Tools then  {display map after each step of the simulation}
             Begin hide; MapTitle.Visible := True;
                   MapTitle.Caption:= 'Map at Year '+IntToStr(Year);
                   UserStop := False;
                   RunPanel.Visible := False;
                   MR := ShowModal;

                   IF (MR=2) then if MessageDlg('Stop Simulation Execution And Return to Main Menu? ',mtconfirmation,[mbyes,mbno],0) = mryes then
                     MR := MRAbort;   // Disambiguation of X key

                   If MR = MRAbort then Begin Result := False; UserStop:=True; exit; End;
                   If MR = MRIgnore then QA_Tools := False;
                   Show; Update; MapTitle.Visible := False; IsVisible := True;
             end;
       ShowRunPanel;

       If Maps_to_MSWord or Maps_to_GIF then Result := CopyMapsToMSWord;
       If Not Result then Begin UserStop := True; Exit; End;

     End;

    UNTIL Year >= MaxYear;

    Save_CSV_File;

{    If (not Display_Screen_Maps) and (not Gridform.IsVisible) then }
    DisposeFile(MapHandle,FALSE,BackupFN);

    { SetLength(Map,0); {deallocate memory}
    ProgForm.Hide;
    Gridform.RunPanel.Visible := False;

  END; {EXECUTESLAMM }

    {------------------------------------------------------------}

Var FixLoop: Integer;

BEGIN {ExecuteRun}
 UserStop := False;

 If (not UncAnalysis) and (not SensAnalysis) then WordInitialized := False;
    Begin           { Loop through Scenarios and Protection Scenarios }
       For IPCC_Scenario := Scen_A1B to Scen_B2 do
        For IPCC_Est := Est_Min to Est_Max do
         For Prot_Scenario := NoProtect to ProtAll do
          If IPCC_Scenarios[IPCC_Scenario] and IPCC_Estimates[IPCC_Est] and Prot_To_Run[Prot_Scenario] then
           If Not ExecuteSLAMM(Prot_Scenario,IPCC_Scenario,IPCC_Est,0) then Exit;

       IPCC_Est := Est_Min;
       IPCC_Scenario := Scen_A1B;
       For Prot_Scenario := NoProtect to ProtAll do
         For FixLoop := 1 to 3 do
           If Fixed_Scenarios[FixLoop] and Prot_To_Run[Prot_Scenario] then
             If Not ExecuteSLAMM(Prot_Scenario,IPCC_Scenario,IPCC_Est,FixLoop) then Exit;

       If RunCustomSLR then
         For Prot_Scenario := NoProtect to ProtAll do
           If Prot_To_Run[Prot_Scenario] then
             If Not ExecuteSLAMM(Prot_Scenario ,IPCC_Scenario,IPCC_Est,9999 {flag for custom}) then Exit;

    End;  {Loop through Scenarios}

    ProgForm.Cleanup;
End;

Constructor TSLAMM_Simulation.Create;
begin
  inherited;

  Changed := True;
  FileN := '';
  SimName := '';
  Descrip := '';
  Site    := TSite.Create;
  TimeStep := 25;
  MaxYear := 2100;
  RunSpecificYears := False;
  YearsString := '';
  NumMMEntries := 0;

  UncertSetup := TSLAMM_Uncertainty.Create;

  Fillchar(Fixed_Scenarios,Sizeof(Fixed_Scenarios),0);
  Fillchar(IPCC_Scenarios,Sizeof(IPCC_Scenarios),0);
  Fillchar(IPCC_Estimates,Sizeof(IPCC_Estimates),0);
  Fillchar(Prot_To_Run,Sizeof(Prot_To_Run),0);

  IPCC_Estimates[Est_Max] := True;
  IPCC_Scenarios[Scen_A1B] := True;
  Prot_To_Run[NoProtect] := True;

  RunCustomSLR := False;
  CustomSLR := 0.59;

  Make_Data_File      := False;
  Display_Screen_Maps := True;
  QA_Tools            := True;
  RunFirstYear        := True;
  Maps_to_MSWord      := False;
  Maps_to_GIF         := False;
  BatchMode           := False;
  SaveGIS             := False;
  SaveElevSGIS        := False;
  SaveSalinityGIS     := False;

  BatchFileN          := '';
  InitZoomFactor      := 1.0;

  SaveMapsToDisk      := False;
  MakeNewDiskMap      := True;

  FWFlows           := nil;
  NumFWFlows        := 0;
  SalRules          := TSalinityRules.Create;

  IncludeDikes        := True;
  OptimizeLevel       := 1;
  ClassicDike        := True;

  SetDefaultCategoryVariables;  {Set to default min and max elevs}

  Fillchar(ElevStats,Sizeof(ElevStats),0);

  GridColors          := DefaultGridColors;

  AssignGridColors2;

  ElevGridColors      := ElevationColors;

   {-- File Name Management}
  ElevFileN := '';
  NWIFileN  := '';

  OutputFileN  := '';

  SLPFileN  := '';
  IMPFileN  := '';
  ROSFileN  := '';
  DikFileN  := '';
  VDFileN   := '';
  UpliftFileN := '';
  Init_ElevStats := False;
  Init_SalStats := False;
  CheckConnectivity := false;
  UseSoilSaturation:= true;
  LoadBlankIfNoElev := true;

   {-------------------------------------------------------------}
  {NO LOAD OR SAVE BELOW}
   SalArray          := nil;
   Map               := nil;
   S60FileN          := '';
   BackupFN          := '';
   DikeInfo          := nil;
   WordApp           := null;
   WordDoc           := null;
   WordInitialized   := False;

   Connect_array:= nil;

  Fillchar(Summary,Sizeof(Summary),0);
  Fillchar(ColLabel,Sizeof(ColLabel),0);
  Fillchar(RowLabel,Sizeof(RowLabel),0);
  FileStarted       := False;
  BMatrix := nil;
  ErodeMatrix := nil;
  MapMatrix := nil;
  DataElev := nil;
  MaxFetchArr := nil;
  RosArray := nil;
  CatSums := nil;
  FRs := nil;
end;

{------------------------------------------------------------------------------}

constructor TSLAMM_Simulation.Load(FN: string);
Var  TS: TFileStream;
begin
  FileN := FN;
  TSText := lowercase(ExtractFileExt(FN)) = '.txt';

  TS := nil;
  Try
    If TSText
      then
        Begin
           AssignFile(GlobalFile,FileN);
           Reset(GlobalFile);
           GlobalLN := 0;
        End
      else TS:=TFileStream.Create(FileN,fmOpenRead);
  Except
    MessageDlg(Exception(ExceptObject).Message,mterror,[mbOK],0);
    Exit;
  End; {Try Except}

  Try
    SetCurrentDir(ExtractFilePath(FileN));
    Load(TS);
  Except
    On E : Exception do
      Begin
       If TS <> nil then TS.Free;
       If TSText then Closefile(GlobalFile);
       ShowMessage('Load Simulation Error: '+E.Message);
      End;
  End;

  If TSText then Closefile(GlobalFile)
            else TS.Free;

  Changed := False;
  SalArray := nil;

end;

Constructor TSLAMM_Simulation.Load(TS: TStream);
Var  ReadVersionNum: Double;
     OLevel, i: Integer;
     AddtlOptimize:Boolean;
     GISEachYear: Boolean;
     GISYrs: String;
     SLR_By_2100: Double; //no longer used

Begin
  GlobalTS := TS;
  TSRead('ReadVersionNum',ReadVersionNum);

  If ReadVersionNum > VersionNum then
    Raise ESLAMMError.Create('File Version Number (' +  FloatToStrF(ReadVersionNum,ffgeneral,6,6)
     + ') is Greater than Executable Version Number ('+ FloatToStrF(VersionNum    ,ffgeneral,6,6)
     + ').  Please update your version of SLAMM.');

  TSRead('SimName',SimName,ReadVersionNum);
  TSRead('Descrip',Descrip,ReadVersionNum);

  Site := TSite.Load(ReadVersionNum,TS);

  TSRead('NumFWFlows',NumFWFlows);
  SetLength(FWFlows,NumFWFlows);
  For i := 0 to NumFWFlows -1 do
    FWFlows[i] := TFWFlow.Load(ReadVersionNum,TS);

  if ReadVersionNum > 6.045 then
    SalRules := TSalinityRules.Load(ReadVersionNum,TS)
  else
    SalRules := TSalinityRules.Create;


  TSRead('TimeStep',TimeStep);
  TSRead('MaxYear',MaxYear);
  If ReadVersionNum < 6.085
    then
      Begin
        RunSpecificYears := False;
        YearsString := '';
      End
    else
      Begin
        TSRead('RunSpecificYears',RunSpecificYears);
        TSRead('YearsString',YearsString,ReadVersionNum);
      End;

  If ReadVersionNum > 6.195 then TSRead('RunUncertainty',RunUncertainty)
                            else RunUncertainty := False;

  If ReadVersionNum > 6.215 then TSRead('RunSensitivity',RunSensitivity)
                            else RunSensitivity := False;


  If TSText then
    Begin
      TSRead('Scen_A1B',IPCC_Scenarios[Scen_A1B]);
      TSRead('Scen_A1T',IPCC_Scenarios[Scen_A1T]);
      TSRead('Scen_A1F1',IPCC_Scenarios[Scen_A1F1]);
      TSRead('Scen_A2',IPCC_Scenarios[Scen_A2]);
      TSRead('Scen_B1',IPCC_Scenarios[Scen_B1]);
      TSRead('Scen_B2',IPCC_Scenarios[Scen_B2]);

      TSRead('Est_Min',IPCC_Estimates[Est_Min]);
      TSRead('Est_Mean',IPCC_Estimates[Est_Mean]);
      TSRead('Est_Max',IPCC_Estimates[Est_Max]);

      TSRead('Fix1.0M',Fixed_Scenarios[1]);
      TSRead('Fix1.5M',Fixed_Scenarios[2]);
      TSRead('Fix2.0M',Fixed_Scenarios[3]);

      TSRead('Prot_To_Run[NoProtect]',Prot_To_Run[NoProtect]);
      TSRead('Prot_To_Run[ProtDeveloped]',Prot_To_Run[ProtDeveloped]);
      TSRead('Prot_To_Run[ProtAll]',Prot_To_Run[ProtAll]);
    End
  else
    Begin
      TS.Read(Fixed_Scenarios,Sizeof(Fixed_Scenarios));
      TS.Read(IPCC_Scenarios,Sizeof(IPCC_Scenarios));
      TS.Read(IPCC_Estimates,Sizeof(IPCC_Estimates));
      TS.Read(Prot_To_Run,Sizeof(Prot_To_Run));
    End;

  If ReadVersionNum > 6.075 then
    Begin
      TSRead('RunCustomSLR',RunCustomSLR);
      TSRead('CustomSLR',CustomSLR);
    End
  else
    Begin
      RunCustomSLR := False;
      CustomSLR := 1.35;
    End;

  TSRead('Make_Data_File',Make_Data_File);
  TSRead('Display_Screen_Maps',Display_Screen_Maps);
  TSRead('QA_Tools',QA_Tools);
  TSRead('RunFirstYear',RunFirstYear);
  TSRead('Maps_to_MSWord',Maps_to_MSWord);

  If ReadVersionNum > 6.275 then TSRead('Maps_to_GIF',Maps_to_GIF)
                            else Maps_to_GIF := False;

  If ReadVersionNum > 6.055 then TSRead('ExtraMaps',ExtraMaps)
                            else ExtraMaps := False;

  If ReadVersionNum > 6.285 then TSRead('DucksMaps',DucksMaps)
                            else DucksMaps := False;

  TSRead('SaveGIS',SaveGIS);
  If ReadVersionNum > 6.235 then
      TSRead('SaveElevsGIS',SaveElevsGIS) else SaveElevsGIS := False;

  If ReadVersionNum > 6.365 then
      TSRead('SaveSalinityGIS',SaveSalinityGIS) else SaveSalinityGIS := False;

  TSRead('BatchMode',BatchMode);

  If ReadVersionNum < 6.085
    then
      Begin
        CheckConnectivity := False;
        UseSoilSaturation := True;
        LoadBlankIfNoElev := True;
      End
    else
      Begin
        TSRead('CheckConnectivity',CheckConnectivity);
        TSRead('UseSoilSaturation',UseSoilSaturation);
        TSRead('LoadBlankIfNoElev',LoadBlankIfNoElev);
      End;

  If ReadVersionNum < 6.095
      then UseBruun := True
      else TSRead('UseBruun',UseBruun);

  If ReadVersionNum<6.195 then
    TSRead('SLR_BY_2100',SLR_BY_2100);

  TSRead('InitZoomFactor',InitZoomFactor);

  TSRead('SaveMapsToDisk',SaveMapsToDisk);
  TSRead('MakeNewDiskMap',MakeNewDiskMap);
  TSRead('IncludeDikes',IncludeDikes);

  OptimizeLevel := 1;
  NumMMEntries := 0;
  If ReadVersionNum < 6.005 then
    Begin
      TSRead('AddtlOptimize',AddtlOptimize);
      If AddtlOptimize then OptimizeLevel := 2;
    End;

  ClassicDike := True;
  if ReadVersionNum > 6.345 then
    Begin
      TSRead('ClassicDike',ClassicDike);
    end;

  If Not TSText then  {elevranges not saved to text}
   If ReadVersionNum > 6.025 then TS.Read(ElevRanges,Sizeof(ElevRanges))
    else Begin
           TS.Read(ElevRanges,414);
           SetDefaultCategoryVariables;
         End;

  If TSText then LoadElevRangesfromText(ElevRanges);

  If Not TSText then  {elevstats not saved to text}
   If ReadVersionNum < 6.295 then
     Begin
        If ReadVersionNum > 6.025
          then
            Begin
              If (ReadVersionNum > 6.195)
                then TS.Read(ElevStats,35264)
                else TS.Read(ElevStats,2192);
            End
          else If ReadVersionNum < 6.015
            Then TS.Read(ElevStats,2216)
            Else TS.Read(ElevStats,1928);
        Fillchar(ElevStats,Sizeof(ElevStats),0);
     End
       else TS.Read(ElevStats,Sizeof(ElevStats));

  If Not TSText then  {gridcolors not saved to text}
    If ReadVersionNum > 6.025
      then TS.Read(GridColors,Sizeof(GridColors))
      else Begin
             TS.Read(GridColors,96);
             GridColors[VegTidalFlat] := $00B7CFB1; {Veg Tidal Flat is grey-green}
             GridColors[BackShore] := $0075D6FF; {backshore is light orange, between dry land and beach}
           End;
  If TSText then  GridColors := DefaultGridColors;

  AssignGridColors2;

  {-- File Name Management}
  TSRead('BatchFileN',BatchFileN,ReadVersionNum);
  TSRead('ElevFileN',ElevFileN,ReadVersionNum);
  TSRead('NWIFileN',NWIFileN,ReadVersionNum);
  TSRead('OutputFileN',OutputFileN,ReadVersionNum);
  TSRead('SLPFileN',SLPFileN,ReadVersionNum);

  TSRead('IMPFileN',IMPFileN,ReadVersionNum);
  TSRead('ROSFileN',ROSFileN,ReadVersionNum);
  TSRead('DikFileN',DikFileN,ReadVersionNum);

  TSRead('VDFileN',VDFileN,ReadVersionNum); {8/26/2009}
  TSRead('UpliftFileN',UpliftFileN,ReadVersionNum);

  If ReadVersionNum >  6.335 then
      TSRead('SalFileN',SalFileN,ReadVersionNum);

  SalFileN := '';  //64-bit release with no salinity maps

  If ReadVersionNum > 6.00 then
    Begin
      If TSText
        then Begin
               TSRead('OptimizeLevel',OLevel);
               OptimizeLevel := Olevel;
             End
        else TS.Read(OptimizeLevel,Sizeof(OptimizeLevel));
      TSRead('NumMMEntries',NumMMEntries);
    End;

  Init_ElevStats := False;
  If ReadVersionNum > 6.015 then
   If Not TSText then TSRead('Init_ElevStats',Init_ElevStats);
  If ReadVersionNum < 6.195 then Init_ElevStats := False;

  Init_SalStats := False;
  If Not TSText then  {SalStats not saved to text}
    If ReadVersionNum > 6.035
      then
        Begin
          If ReadVersionNum < 6.305
            then TS.Read(SalStats,36504)
            else TS.Read(SalStats,Sizeof(SalStats));
          TSRead('Init_SalStats',Init_SalStats);
        End;

  If ReadVersionNum > 6.065
    then Begin
           TSRead('GISEachYear',GISEachYear);
           TSRead('GISYears',GISYrs,ReadVersionNum);
           GISForm.WriteEachYear.Checked := GISEachYear;
           GISForm.ChooseYears.Checked := not GISEachYear;

           GISForm.Edit1.Text := GISYrs;
         End
    else GISForm.WriteEachYear.Checked := True;

  If ReadVersionNum > 6.199 then UncertSetup := TSLAMM_Uncertainty.Load(ReadVersionNum, TS, Self)
                            else UncertSetup := TSLAMM_Uncertainty.Create;

  If TSText then TSRead('ExecuteImmediately',DRIVE_EXTERNALLY);

  Changed := False;

  {-------------------------------------------------------------}
  {NO LOAD OR SAVE BELOW}

   Map               := nil;
   S60FileN          := '';
   BackupFN          := '';
   DikeInfo          := nil;
   WordApp           := null;
   WordDoc           := null;
   WordInitialized   := False;

   Connect_array:= nil;

  Fillchar(Summary,Sizeof(Summary),0);
  Fillchar(ColLabel,Sizeof(ColLabel),0);
  Fillchar(RowLabel,Sizeof(RowLabel),0);
  FileStarted       := False;
  BMatrix := nil;
  ErodeMatrix := nil;
  MapMatrix := nil;
  DataElev := nil;
  MaxFetchArr := nil;
  RosArray := nil;
  CatSums := nil;
  FRs := nil;
  SalArray := nil;

end;
{------------------------------------------------------------------------------}

procedure TSLAMM_Simulation.Save(FN: string);
Var  TS: TFileStream;
begin
  FileN := FN;

  TSText := lowercase(ExtractFileExt(FN)) = '.txt';

  TS := nil;
   Try
     If TSText
       then
         Begin
           AssignFile(GlobalFile,FileN);
           Rewrite(GlobalFile);
           GlobalLN := 0;
         End
       else TS:=TFileStream.Create(FileN,fmCreate);
    Except
      MessageDlg(Exception(ExceptObject).Message,mterror,[mbOK],0);
      Exit;
    End; {Try Except}

  Save(TStream(TS));

  If TSText then Closefile(GlobalFile)
            else TS.Free;
End;

Procedure TSLAMM_Simulation.Save(var TS:TStream);
Var  OLevel, i: Integer;
     GISEachYear: Boolean;
     GISYears: String;

Begin
  GlobalTS := TS;
  TSWrite('ReadVersionNum',VersionNum);

  TSWrite('SimName',SimName);
  TSWrite('Descrip',Descrip);

  Site.Store(TStream(TS));

  TSWrite('NumFWFlows',NumFWFlows);
  For i := 0 to NumFWFlows-1 do
    FWFlows[i].Store(TStream(TS));

  SalRules.Store(TS);

  TSWrite('TimeStep',TimeStep);
  TSWrite('MaxYear',MaxYear);
  TSWrite('RunSpecificYears',RunSpecificYears);
  TSWrite('YearsString',YearsString);
  TSWrite('RunUncertainty',RunUncertainty);
  TSWrite('RunSensitivity',RunSensitivity);

  If TSText then
    Begin
      TSWrite('Scen_A1B',IPCC_Scenarios[Scen_A1B]);
      TSWrite('Scen_A1T',IPCC_Scenarios[Scen_A1T]);
      TSWrite('Scen_A1F1',IPCC_Scenarios[Scen_A1F1]);
      TSWrite('Scen_A2',IPCC_Scenarios[Scen_A2]);
      TSWrite('Scen_B1',IPCC_Scenarios[Scen_B1]);
      TSWrite('Scen_B2',IPCC_Scenarios[Scen_B2]);

      TSWrite('Est_Min',IPCC_Estimates[Est_Min]);
      TSWrite('Est_Mean',IPCC_Estimates[Est_Mean]);
      TSWrite('Est_Max',IPCC_Estimates[Est_Max]);

      TSWrite('Fix1.0M',Fixed_Scenarios[1]);
      TSWrite('Fix1.5M',Fixed_Scenarios[2]);
      TSWrite('Fix2.0M',Fixed_Scenarios[3]);

      TSWrite('Prot_To_Run[NoProtect]',Prot_To_Run[NoProtect]);
      TSWrite('Prot_To_Run[ProtDeveloped]',Prot_To_Run[ProtDeveloped]);
      TSWrite('Prot_To_Run[ProtAll]',Prot_To_Run[ProtAll]);
    End
  else
    Begin
      TS.Write(Fixed_Scenarios,Sizeof(Fixed_Scenarios));
      TS.Write(IPCC_Scenarios,Sizeof(IPCC_Scenarios));
      TS.Write(IPCC_Estimates,Sizeof(IPCC_Estimates));
      TS.Write(Prot_To_Run,Sizeof(Prot_To_Run));
    End;

  TSWrite('RunCustomSLR',RunCustomSLR);
  TSWrite('CustomSLR',CustomSLR);

  TSWrite('Make_Data_File',Make_Data_File);
  TSWrite('Display_Screen_Maps',Display_Screen_Maps);
  TSWrite('QA_Tools',QA_Tools);
  TSWrite('RunFirstYear',RunFirstYear);
  TSWrite('Maps_to_MSWord',Maps_to_MSWord);
  TSWrite('Maps_to_GIF',Maps_to_GIF);
  TSWrite('ExtraMaps',ExtraMaps);
  TSWrite('DucksMaps',DucksMaps);

  TSWrite('SaveGIS',SaveGIS);
  TSWrite('SaveElevsGIS',SaveElevsGIS);
  TSWrite('SaveSalinityGIS',SaveSalinityGIS);

  TSWrite('BatchMode',BatchMode);

  TSWrite('CheckConnectivity',CheckConnectivity);
  TSWrite('UseSoilSaturation',UseSoilSaturation);
  TSWrite('LoadBlankIfNoElev',LoadBlankIfNoElev);

  TSWrite('UseBruun',UseBruun);

  TSWrite('InitZoomFactor',InitZoomFactor);

  TSWrite('SaveMapsToDisk',SaveMapsToDisk);
  TSWrite('MakeNewDiskMap',MakeNewDiskMap);
  TSWrite('IncludeDikes',IncludeDikes);
  TSWrite('ClassicDike',ClassicDike);

  If Not TSText
    then TS.Write(ElevRanges,Sizeof(ElevRanges))
    else SaveElevRangesToText(ElevRanges);

  If Not TSText then  {elevstats not saved to text}
    TS.Write(ElevStats,Sizeof(ElevStats));

  If Not TSText then  {grid colors not saved to text}
    TS.Write(GridColors,Sizeof(GridColors));

  {-- File Name Management}
  TSWrite('BatchFileN',BatchFileN);
  TSWrite('ElevFileN',ElevFileN);
  TSWrite('NWIFileN',NWIFileN);
  TSWrite('OutputFileN',OutputFileN);
  TSWrite('SLPFileN',SLPFileN);

  TSWrite('IMPFileN',IMPFileN);
  
  TSWrite('ROSFileN',ROSFileN);

  TSWrite('DikFileN',DikFileN);

  TSWrite('VDFileN',VDFileN);
  TSWrite('UpliftFileN',UpliftFileN);

  TSWrite('SalFileN',SalFileN);  //new to 6.34

  If TSText
    then Begin
           OLevel := OptimizeLevel;
           TSWrite('OptimizeLevel',OLevel);
         End
       else TS.Write(OptimizeLevel,Sizeof(OptimizeLevel));
  TSWrite('NumMMEntries',NumMMEntries);

  If Not TSText then  {Elev & Sal Stats not saved to text}
    Begin
      TSWrite('Init_ElevStats',Init_ElevStats);
      TS.Write(SalStats,Sizeof(SalStats));
      TSWrite('Init_SalStats',Init_SalStats);
    End;

  GISEachYear := GISForm.WriteEachYear.Checked ;
  GISYears    := GISForm.Edit1.Text;
  TSWrite('GISEachYear',GISEachYear);
  TSWrite('GISYears',GISYears);

  UncertSetup.Store(TS);

  If TSText then TSWrite('ExecuteImmediately',FALSE);  {default is not to start simulations automatically}
  Changed := False;

end;

{------------------------------------------------------------------------------}

Destructor TSLAMM_Simulation.Destroy;
Begin
  Wordapp := null;
  Worddoc := null;
  If Site<>nil then Site.Free;
  BMatrix := nil;
  MapMatrix:= nil;
  If SalArray<>nil then SalArray.Destroy;

  UncertSetup.Free;
  SalRules.Free;

  DisposeMem;

  inherited;
End;

{-----------------------------------------------------------------------------------}

Procedure TSLAMM_Simulation.AssignGridColors2;
Var Cat: Category;
Begin
  for Cat := FirstCategory to LastCategory do
    Begin
      GridColors2[Cat].rgbtRed := Byte(GridColors[Cat]);
      GridColors2[Cat].rgbtGreen := Byte(GridColors[Cat] shr 8);
      GridColors2[Cat].rgbtBlue := Byte(GridColors[Cat] shr 16);
    End;
End;

End.


